import importlib
from abc import ABC
from typing import Any, Dict, Optional

from pyspark.sql import DataFrame, SparkSession

from karadoc.common.class_utils import find_class_from_module
from karadoc.common.conf import ConfBox, get_connection_conf
from karadoc.common.conf.configurable_class import ConfigurableClass, ConfParam


class Connector(ABC):
    """Users can implement this class to write their own custom connectors.

    Implementations may override the following methods:

    - `read(self, source: Dict) -> DataFrame` : reads from an external source and return a Spark DataFrame (batch)
    - `write(self, df: DataFrame, dest: Dict)` : writes a given DataFrame to an external destination (batch)
    - `read_stream(self, source: Dict) -> DataFrame` : reads from an external streaming source and
      return a Spark DataFrame
    - `write_stream(self, df: DataFrame, dest: Dict)`  : writes a given streaming DataFrame to an external sink

    They can also have their own public methods for side-effects.
    If you want to call a connector's method with side-effects, it is recommended to NOT do it in the ACTION_FILE's
    `run`/`stream` method, but instead by overriding on of the following methods:

    - `read_external_input(source: Dict) -> DataFrame` : read a single external input and return a DataFrame
    - `write_external_output(df: DataFrame, dest: Dict) -> DataFrame` : write a DataFrame to a single external output

    This will make regression-testing easier.

    In some very specific cases, one might want to read from multiple sources or write to multiple destinations
    by having shared side-effects between them (for instance, if you want to read two DataFrames from a single
    source with a very heavy initialization cost, or if you want to write to two destination and rollback the first
    if the second fails). In such case, you might want to overwrite the following methods instead:

    - `read_external_inputs() -> Dict[str, DataFrame]` : read all external inputs declared in this job and
      return a Dict(alias, DataFrame)
    - `write_external_outputs(df: DataFrame): Write a DataFrame to all external outputs declared in
      this job (not available in streaming)
    """

    spark: SparkSession = None
    conf: ConfBox = None

    def __init__(self, spark: SparkSession, conf: ConfBox) -> None:
        """Instantiates the `Connector` base abstract class

        :param spark: The Spark Session
        :param conf: A DynaBox or a dict-like object
        """
        self.conf = conf
        self.spark = spark

    def write(self, df: DataFrame, dest: Dict[str, Any]):
        """Write the given DataFrame to an external output described by `dest`

        :param df: the DataFrame to export
        :param dest: A dictionary describing the destination
        :return:
        """
        raise NotImplementedError()

    def write_stream(self, df: DataFrame, dest: Dict[str, Any]):
        """Write the given streaming DataFrame to an external output described by `dest`

        :param df:
        :param dest:
        :return:
        """
        raise NotImplementedError()

    def read(self, source: Dict[str, Any]) -> DataFrame:
        """Write the given external input described by `source` and return it as a DataFrame

        :param source:
        :return:
        """
        raise NotImplementedError()

    def read_stream(self, source: Dict[str, Any]) -> DataFrame:
        """Write the given external input described by `source` and return it as a streaming DataFrame

        :param source:
        :return:
        """
        raise NotImplementedError()

    def test_connection(self):
        """Test that the connection works without reading or writing anything
        (or by writing a test output and deleting it afterwards, depending on the implementation)"""
        return NotImplemented


class ConfigurableConnector(ConfigurableClass, Connector, ABC):
    """Improved version of the :class:`Connector` class, that also extends :class:`ConfigurableClass`.

    Connectors inheriting from this class will benefit from automatic parameter validation.
    """

    type = ConfParam(type=str, required=True, description="Type of the connector")

    disable = ConfParam(
        type=bool,
        required=False,
        default=False,
        description="Set to true to disable this connector. Only disabled jobs may reference disabled connectors.",
    )

    def __init__(self, spark: SparkSession, conf: ConfBox):
        ConfigurableClass.__init__(self, conf)
        Connector.__init__(self, spark, conf)


def load_connector(connection_name: str, spark: SparkSession) -> Connector:
    """Load a connector configured with the given connection.

    This method will look in the settings.toml for an entry that looks like this:

    [{current_env}.connection.{connection_name}]
      type = {module_name}
      {connection_setting_key_1} = {connection_setting_value_1}
      ...

    It will then load the specified module which should define a class that inherits from
    :class:`karadoc.common.connector.Connector` and build a Connector of that type,
    configured with the {connection_settings} found in the settings.toml

    :param connection_name: Name of a connection defined in settings.toml
    :param spark: a Spark Session
    :return:
    """
    return _load_connector_for_env(connection_name, spark, None)


def _load_connector_for_env(connection_name: str, spark: SparkSession, env: Optional[str]) -> Connector:
    """Load a connector for the specified connection name in the specified dynaconf environment.

    This method will look in the settings.toml for an entry that looks like this:

    [{env}.connection.{connection_name}]
      type = {module_name}
      {connection_setting_key_1} = {connection_setting_value_1}
      ...

    It will then load the specified module which should define a class that inherits from
    :class:`karadoc.common.connector.Connector` and build a Connector of that type,
    configured with the {connection_settings} found in the settings.toml

    Warning: If two implementations of Connector are defined in the same module, karadoc will not be able to know which
    one to pick. Therefore each implementation of Connector must be kept in a separate module.

    :param connection_name: Name of a connection defined in settings.toml
    :param spark: a Spark Session
    :param env: The environment on which to load the connection
        (useful for running validations on multiple environments)
    :return:
    """
    connection_conf = get_connection_conf(connection_name, env)
    connector_mock = connection_conf.get("mock")
    if connector_mock is not None:
        module_name = connector_mock
    else:
        module_name = connection_conf["type"]
    module = importlib.import_module(module_name)
    connector_class = find_class_from_module(module, Connector)
    return connector_class(spark, connection_conf)



import os
import zipfile
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import List, Union

import karadoc
from karadoc.common import conf
from karadoc.common.commands.command import Command

CONF_TARGET = "conf"
LIBS_TARGET = "libs"
MODEL_TARGET = "model"


def __write_to_zip(zip_file, source_path, dest_path, logger):
    """Add the file at the specified 'path' to the specified 'zip_file'"""
    zip_file.write(source_path, dest_path)
    if logger is not None:
        logger.info(f"adding {source_path} as {dest_path}")


def _add_folder_to_zip_file(input_dir: str, output_dir: str, zip_file: zipfile.ZipFile, logger=None):
    """Add the content of 'input_dir' to the 'zip_file' archive and ignore '__pycache__' folders"""
    for dirpath, dirnames, filenames in os.walk(input_dir):
        relpath = os.path.relpath(dirpath, input_dir)
        for name in sorted(dirnames):
            source_path = os.path.normpath(os.path.join(dirpath, name))
            dest_path = os.path.normpath(os.path.join(output_dir, relpath, name))
            if "__pycache__" not in name:
                __write_to_zip(zip_file, source_path, dest_path, logger)
        if "__pycache__" not in relpath:
            for name in filenames:
                source_path = os.path.normpath(os.path.join(dirpath, name))
                dest_path = os.path.normpath(os.path.join(output_dir, relpath, name))
                if os.path.isfile(source_path):
                    __write_to_zip(zip_file, source_path, dest_path, logger)


def _add_folder_to_python_zip_file(input_dir, zip_file: zipfile.PyZipFile, logger=None):
    """Add the content of 'input_dir' to the 'zip_file' archive."""
    zip_file.writepy(input_dir)
    if logger is not None:
        logger.info("adding '%s' to '%s'", input_dir, zip_file)


def make_zipfile(output_base_name: str,
                 input_dirs: List[str],
                 keep_root_folder: bool,
                 dry_run=False,
                 logger=None):
    """Create a zip file from all the files in the 'input_dirs'.

    The output zip file will be named 'output_base_name' + ".zip".
    Return the name of the output zip file.

    :param output_base_name:
    :param input_dirs:
    :param keep_root_folder: set to True if you want to keep the root folders in the zip file
    :param dry_run: set to True to perform a dry-run
    :param logger: pass logger
    :return:
    """

    zip_filename = output_base_name + ".zip"
    archive_dir = os.path.dirname(output_base_name)

    if archive_dir and not os.path.exists(archive_dir):
        if logger is not None:
            logger.info(f"creating {archive_dir}")
        if not dry_run:
            os.makedirs(archive_dir)

    if logger is not None:
        logger.info(f"creating '{zip_filename}' and adding '{input_dirs}' to it")

    zip_file_type = zipfile.ZipFile

    if not dry_run:
        with zip_file_type(zip_filename, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            for input_dir in input_dirs:
                if keep_root_folder:
                    output_dir = str(Path(input_dir).name)
                else:
                    output_dir = ""
                _add_folder_to_zip_file(input_dir, output_dir, zf, logger)

    return zip_filename


class PackageCommand(Command):
    description = "(Experimental) create deployable archives in the target/ folder"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            '--target',
            dest='target',
            type=str,
            default='target',
            metavar='FOLDER',
            help='Write archives in target FOLDER'
        )

    @staticmethod
    def do_command(args: Namespace):
        target_folder = Path(args.target)

        karadoc_code_dir = str(Path(karadoc.__file__).parent)
        libs_folder = conf.get_libs_folder_location()
        libs = [str(Path(libs_folder) / f) for f in os.listdir(libs_folder)]
        input_dirs = [karadoc_code_dir] + libs
        PackageCommand.zip_folders(input_dirs=input_dirs,
                                   target_folder=target_folder / LIBS_TARGET,
                                   keep_root_folder=True)

        PackageCommand.zip_folders(input_dirs=conf.get_model_folder_location(),
                                   target_folder=target_folder / MODEL_TARGET,
                                   keep_root_folder=False)

        PackageCommand.zip_folders(input_dirs=conf.get_conf_folder_location(),
                                   target_folder=target_folder / CONF_TARGET,
                                   keep_root_folder=False)

    @staticmethod
    def zip_folders(input_dirs: Union[str, List[str]], target_folder: Union[str, Path], keep_root_folder: bool):
        """Create a zip file from all the files in the 'input_dirs'.

        Example
        let's concider the folder input_dir :
                input_dir/
                └── sub_dir
             zip_folders('input_dir', 'target', keep_root_folder=True)
             generates a target.zip containing :
                input_dir/
                   └── sub_dir
             zip_folders('input_dir', 'target', keep_root_folder=False)
             generates target.zip containing :
                sub_dir

        :param target_folder: the output zip file will be named 'target_folder' + ".zip".
        :param input_dirs: folders to zip
        :param keep_root_folder: set to True if you want to keep the root folders in the zip file
        """
        if isinstance(input_dirs, str):
            input_dirs = [input_dirs]
        output_zip = make_zipfile(output_base_name=str(target_folder),
                                  input_dirs=input_dirs,
                                  keep_root_folder=keep_root_folder)
        modules = ", ".join(input_dirs)
        print(f"Created '{output_zip}' containing the following modules: {modules}")



from abc import ABC
from argparse import ArgumentParser

from karadoc.common.commands.command import Command


class SparkConfOption(Command, ABC):
    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        from karadoc.common.commands.parsing import StoreDictKeyValue

        parser.add_argument(
            "--spark-conf",
            nargs="*",
            dest="spark_conf",
            metavar="KEY=VALUE",
            required=False,
            action=StoreDictKeyValue,
            help="Specify Spark configuration parameters. Use this only in LOCAL mode. "
            "In distributed mode, please pass them via the spark-submit command.",
        )



from typing import Any


def assert_true(assertion: bool, error_message: Any = None) -> None:
    """Raise an AssertionError with the given error_message if the assertion passed is false

    >>> assert_true(3==4, "3 <> 4")
    Traceback (most recent call last):
    ...
    AssertionError: 3 <> 4

    >>> assert_true(3==3, "3 <> 4")

    :param assertion: assertion that will be checked
    :param error_message: error message to display if the assertion is false
    """
    if not assertion:
        if error_message is None:
            raise AssertionError()
        else:
            raise AssertionError(str(error_message))



from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Dict, Optional, Type

from karadoc.common.commands.command import Command
from karadoc.common.commands.exec import (
    add_commands_to_parser,
    do_command,
    load_commands_from_package,
)
from karadoc.common.commands.return_code import ReturnCode


class CommandGroup(Command, ABC):
    """A group of commands.

    Command groups are defined via a python module with the following structure:
    - An __init__.py file containing an implementation of CommandGroup
    - One python file per command in the group each one containing an implementation of Command

    A command group can also contain other command groups.

    A CommandGroup implementation must implement the following element:

    - description: str.

    Example of command group structure
    ----------------------------------

    The example file structure below will define the following commands
    - karadoc build validate
    - karadoc build package

        # commands/build/__init__.py :
        class BuildCommandGroup(CommandGroup):
            description = "set of commands used for code validation and packaging"

        # commands/build/validate.py :
        class ValidateCommand(Command):
            description = "apply validation checks on all tables and connections referenced in the project"

            @staticmethod
            def add_arguments(parser: ArgumentParser):
                ...

            @staticmethod
            def do_command(args: Namespace) -> Optional[ReturnCode]:
                ...

        # commands/build/package.py :
        class PackageCommand(Command):
            description = "create a deployable package of the project"

            @staticmethod
            def add_arguments(parser: ArgumentParser):
                ...

            @staticmethod
            def do_command(args: Namespace) -> Optional[ReturnCode]:
                ...
    """

    parser: ArgumentParser
    sub_commands_by_name: Dict[str, Type[Command]]

    @property
    @abstractmethod
    def description(self) -> str:
        """Description of the command group which will be displayed in the help"""
        pass

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        cls.parser = parser
        cls.sub_commands_by_name = load_commands_from_package(
            cls._get_command_module(), command_depth=cls.command_depth + 1
        )
        add_commands_to_parser(parser, cls.sub_commands_by_name, command_depth=cls.command_depth + 1)

    @classmethod
    def do_command(cls, args: Namespace) -> Optional[ReturnCode]:
        return do_command(args, cls.parser, cls.sub_commands_by_name, command_depth=cls.command_depth + 1)



from pyspark.sql import DataFrame, SparkSession

from karadoc.common.conf import APPLICATION_NAME, ConfBox
from karadoc.common.conf.configurable_class import ConfParam
from karadoc.common.connector import ConfigurableConnector


class JdbcConnector(ConfigurableConnector):
    """This connector leverages `Spark's jdbc <https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html>`_
    connector.

    It builds the following JDBC url from its configuration parameters:
    "jdbc:{protocol}://{host};user={user};password={password};database={database}"

    It requires to have the jars of the corresponding jdbc protocol available in the path of the Spark JVMs.
    """

    protocol = ConfParam(
        required=True,
        type=str,
        description="JDBC protocol to use (e.g. sqlserver, psql)."
        "A jar with the corresponding protocol implementation must be available in Spark's JVM path.",
    )
    host = ConfParam(
        required=True, type=str, description="Hostname of the SQL server (e.g. example.database.windows.net)."
    )
    user = ConfParam(required=True, type=str, description="Name of the SQL user.")
    password = ConfParam(
        required=True, type=str, secret=True, description="Password (should be stored in a secret vault)."
    )
    database = ConfParam(required=True, type=str, description="SQL database to connect to.")

    def __init__(self, spark: SparkSession, connection_conf: ConfBox):
        super().__init__(spark, connection_conf)

    def _get_url(self):
        protocol = self.protocol.get()
        host = self.host.get()
        user = self.user.get()
        password = self.password.get()
        database = self.database.get()
        url = (
            f"jdbc:{protocol}://{host}"
            f";user={user};password={password};database={database};applicationName={APPLICATION_NAME}"
        )
        return url

    def read(self, source: dict):
        """Read data from an external input via
        `Spark's jdbc <https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html>`_ connection.

        If the input table has column that contains dots ('.'), these are
        automatically replaced by underscores to avoid bugs downstream.

        The source must have the following attributes:

        - "connection": name of the jdbc connection
        - "table": name of the table
        - "custom_query": an optional custom query that can be use instead of "table" for advanced use.
        The query must be written as a sub-select (ex: "(SELECT colA from table) as T")

        The corresponding jdbc connection must have the following attributes:

        - type = 'jdbc'
        - protocol = 'sqlserver'
        - host = Server name
        - user = User name
        - password = Password
        - database = Database name

        Known limitations:

        - When using the "table" attribute automatically adds a "WITH (NOLOCK)" hint, which is not compatible
        with Azure Datawarehouse's external tables.

        :param source: a dictionary describing the source
        :return: a DataFrame
        """
        sql_url = self._get_url() + ";ApplicationIntent=ReadOnly"
        custom_query = source.get("custom_query")
        if custom_query is not None:
            df = self.spark.read.jdbc(sql_url, custom_query)
        else:
            df = (
                self.spark.read.format("jdbc")
                .option("url", sql_url)
                .option("dbtable", source["table"])
                .option("isolationLevel", "READ_UNCOMMITTED")
                .load()
            )
        # Replace dots with underscore in column names
        df = df.toDF(*[c.replace(".", "_") for c in df.columns])
        return df

    def write(self, df: DataFrame, dest) -> None:
        """Write dataframe to an external output via
        `Spark's jdbc <https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html>`_ connection.

        The dest must have the following attributes:
        - "connection": name of the jdbc connection
        - "table": name of the table
        - "options": (optional) extra options to pass to the jdbc connector write method
        # https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html

        The corresponding jdbc connection must have the following attributes:

        - type = 'jdbc'
        - protocol = 'sqlserver'
        - host = Server name
        - user = User name
        - password = Password
        - database = Database name

        :param df: the DataFrame to export
        :param dest: a dictionary describing the destination
        :return:
        """
        sql_url = self._get_url() + ";ApplicationIntent=ReadWrite"
        df.write.options(**dest.get("options", {})).jdbc(sql_url, dest["table"], dest["mode"])

    def test_connection(self):
        sql_url = self._get_url() + ";ApplicationIntent=ReadOnly"
        df = self.spark.read.jdbc(sql_url, "(SELECT 1 as ok) temp")
        if df.collect()[0].ok == 1:
            return True
        else:
            df.show()
            return False



import copy
from dataclasses import dataclass
from typing import Mapping, Optional

from karadoc.common.observability.logging_context import (
    LoggingContext,
    get_logging_context,
)


@dataclass(init=False)
class LogEvent:
    """Structured logging data"""

    message: str
    """Human-readable message"""
    event_type: str
    """Event type, useful for grouping and filtering"""
    extra: Optional[Mapping[str, str]]
    """Additional information specific to this event type"""
    context: LoggingContext
    """Static thread-local context information modified by the system"""

    def __init__(self, message: str, event_type: str, extra: Optional[Mapping[str, str]] = None):
        self.message = message
        self.event_type = event_type
        self.extra = extra
        self.context = copy.deepcopy(get_logging_context())



import re
from typing import Optional, Tuple

filter_parsing_pattern = re.compile(r"""(?:(\d*)([+]))?([^+]+)(?:([+])(\d*))?""")

s = "+schema.table+2"


def parse_filter_string(filter_str: str) -> Tuple[str, int, int]:
    """Parse a graph filter string and return a tuple (model_name, upstream_depth, downstream_depth)

    A depth of 0 or less means no exploration in that direction
    A depth equal to None means unbounded exploration in that direction

    >>> parse_filter_string("schema.table")
    ('schema.table', 0, 0)
    >>> parse_filter_string("schema.table+")
    ('schema.table', 0, None)
    >>> parse_filter_string("+schema.table")
    ('schema.table', None, 0)
    >>> parse_filter_string("+schema.table+")
    ('schema.table', None, None)
    >>> parse_filter_string("+schema.table+2")
    ('schema.table', None, 2)
    >>> parse_filter_string("3+schema.table+2")
    ('schema.table', 3, 2)

    :param filter_str:
    :return: a tuple (model_name, upstream_depth, downstream_depth)
    """
    groups = filter_parsing_pattern.match(filter_str).groups()

    def get_exploration_depth(is_enabled: bool, str_depth: Optional[str]) -> Optional[int]:
        """Return the parsed value of exploration depth.

        - "" means no exploration -> 0
        - "+" means unbounded exploration -> None
        - "+N" means exploration bounded to N -> N
        """
        if is_enabled:
            if str_depth:
                return int(str_depth)
            else:
                return None
        else:
            return 0

    if len(groups) == 5:
        (upstream_depth, is_upstream, table_name, is_downstream, downstream_depth) = groups
        upstream_depth = get_exploration_depth(is_upstream, upstream_depth)
        downstream_depth = get_exploration_depth(is_downstream, downstream_depth)
        return table_name, upstream_depth, downstream_depth
    else:
        raise ValueError(f"Could not parse the graph filter {s}")


class GraphFilter:
    node: str
    upstream_depth: Optional[int]
    downstream_depth: Optional[int]

    def __init__(self, filter: str, default_upstream_depth: Optional[int], default_downstream_depth: Optional[int]):
        self.node, self.upstream_depth, self.downstream_depth = parse_filter_string(filter)
        if self.upstream_depth == 0:
            self.upstream_depth = default_upstream_depth
        if self.downstream_depth == 0:
            self.downstream_depth = default_downstream_depth



class HasDisable:
    def __init__(self) -> None:
        self.disable = False
        """Set to true to make the job fail if it is executed. Keeping disabled jobs instead of deleting them is useful
        to keep data lineage of your legacy data while making sure the job won't be accidentally run. Disabled jobs will
        be colored in grey in the dependency graph generated by the `show_graph` command."""



from argparse import ArgumentParser, Namespace

from karadoc.common.commands.command import Command


class ShowAirflowDagGraphCommand(Command):
    description = "display the dependency graph of all airflow dags"

    @staticmethod
    def add_arguments(parser: ArgumentParser) -> None:
        # This command takes no arguments
        pass

    @staticmethod
    def do_command(args: Namespace) -> None:
        from karadoc.airflow.dag_graph import build_dag_bag

        dag_bag = build_dag_bag()

        from karadoc.airflow.dag_graph import (
            build_dag_graph,
            build_graph,
            render_dag_graph,
        )

        graph = build_graph(dag_bag)
        dag_graph = build_dag_graph(graph)
        render_dag_graph(dag_graph)



from typing import Dict, Union

from pyspark.sql import DataFrame

from karadoc.common.conf import CONNECTION_GROUP
from karadoc.common.job_core.has_spark import HasSpark


def _read_stream_external_input_signature_check(source: Dict) -> DataFrame:
    """We uses this empty method when to check the signature of the equivalent method defined in the STREAM files"""
    pass


def _read_stream_external_inputs_signature_check() -> Dict[str, DataFrame]:
    """We uses this empty method when to check the signature of the equivalent method defined in the STREAM files"""
    pass


class HasStreamExternalInputs(HasSpark):
    def __init__(self) -> None:
        super().__init__()
        # Private attributes
        self.__read_external_input = None
        self.__read_external_inputs = None

        # Attributes that the user may change in action files
        self.external_inputs: Dict[str, dict] = {}
        """Use this to declare your job's external inputs.
        Expected format is a dictionary of (table alias, table declaration)
        The table description must be a dict specifying the connection to use and all additional parameters specific
        to the external input. The corresponding connection must be properly configured as described in the connector's
        documentation.

        Example:
        {
            "external_input_alias": {
                "connection": "name_of_the_connection_used",
                "source_param_name": "source_param_value"
            }
        }
        """

    def _read_external_input_default(self, source: Dict) -> DataFrame:
        connector = self.get_input_connector(source)
        return connector.read_stream(source)

    def _read_external_inputs_default(self) -> Dict[str, DataFrame]:
        return {source_alias: self.read_external_input(source_alias) for source_alias in self.external_inputs}

    def read_external_input(self, source_alias: str) -> DataFrame:
        """Reads a given external input and returns it as a Spark DataFrame

        :param source_alias: the alias of the source in job.external_inputs
        :return: a DataFrame
        """
        source = self.external_inputs[source_alias]
        if self.__read_external_input is None:
            return self._read_external_input_default(source)
        else:
            return self.__read_external_input(source)

    def read_external_inputs(self) -> Dict[str, DataFrame]:
        """Reads all declared external inputs and returns them as Spark DataFrames

        :return: a Dict[alias, DataFrame]
        """
        if self.__read_external_inputs is None:
            return self._read_external_inputs_default()
        else:
            return self.__read_external_inputs()

    def load_external_inputs_as_views(self, cache=False):
        for alias in self.external_inputs:
            df = self.read_external_input(alias)
            if cache:
                df = df.cache()
            df.createOrReplaceTempView(alias)

    def get_input_connector(self, source: Union[str, Dict]):
        if type(source) == str:
            source = self.external_inputs[source]
        from karadoc.common.connector import load_connector

        return load_connector(source[CONNECTION_GROUP], self.spark)



from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

from pyspark.sql import DataFrame

from karadoc.common import conf
from karadoc.common.job_core.has_spark import HasSpark, _table_name_to_hdfs_path
from karadoc.common.table_utils import parse_table_name


def _output_partition_to_dynamic_partitions(partition) -> List[str]:
    dynamic_partitions = []
    if partition:
        for element in partition:
            if element and type(element) == str:
                dynamic_partitions.append(element)
    return dynamic_partitions


class HasOutput(HasSpark, ABC):
    def __init__(self) -> None:
        # Attributes that the user may change
        self.output_partition: List[Union[Tuple[str, ...], str]] = []
        self.output_mode = "OVERWRITE"
        self.output_options = {}

        # Attributes that the user is not supposed to change
        self.output: Optional[str] = None
        self.output_warehouse_dir = conf.get_warehouse_folder_location()

    @property
    def output_partitioning_type(self) -> Optional[str]:
        """Indicates the type of partitioning of the output.
        Possible return values are:
        - None: no partitioning
        - "static": dynamic partitioning
        - "dynamic": static partitioning

        :return: either None, "static", or "dynamic"
        """
        if self.output_partition:
            if type(self.output_partition[0]) == str:
                return "dynamic"
            else:
                return "static"
        else:
            return None

    @property
    def output_partition_names(self) -> List[str]:
        """Returns the names of the output partitions.
        Works whether dynamic or static partitioning is used.

        :return: a list of string
        """
        output_partitioning_type = self.output_partitioning_type
        if output_partitioning_type == "dynamic":
            return [p_name for p_name in self.output_partition]
        elif output_partitioning_type == "static":
            return [p_name for p_name, p_value in self.output_partition]
        else:
            return []

    @abstractmethod
    def write_table(self, df: DataFrame):
        pass

    def hdfs_output(self) -> str:
        (schema_name, table_name, _) = parse_table_name(self.output)
        return _table_name_to_hdfs_path(self.output_warehouse_dir, schema_name, table_name, self.output_partition)

    def dynamic_partitions(self) -> List[str]:
        return _output_partition_to_dynamic_partitions(self.output_partition)



import traceback
from argparse import ArgumentParser, Namespace

from termcolor import colored

from karadoc.common.commands.command import Command
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.conf import CONNECTION_GROUP, get_conn_conf_id
from karadoc.common.connector import load_connector


def check_connection(conn_name, job):
    return_code = ReturnCode.Success
    conn_conf_id = get_conn_conf_id(conn_name, group=CONNECTION_GROUP)
    try:
        connector = load_connector(conn_name, job.spark)
        disabled = connector.conf.get("disable", False)
        if disabled:
            print(colored(f"{conn_conf_id} ({connector.conf.type}): disabled", "white"))
        else:
            test_result = connector.test_connection()
            if test_result == NotImplemented:
                print(colored(f"{conn_conf_id} ({connector.conf.type}): not implemented", "yellow"))
            elif test_result:
                print(colored(f"{conn_conf_id} ({connector.conf.type}): ok", "green"))
            else:
                print(colored(f"{conn_conf_id} ({connector.conf.type}): not ok", "red"))
                return_code = ReturnCode.Error
    except Exception:
        traceback.print_exc()
        return_code = ReturnCode.Error
        print(colored(f"{conn_conf_id}: not ok", "red"))
    return return_code


def check_connections(args) -> ReturnCode:
    from karadoc.common import Job, conf

    job = Job()
    init_job(job=job, raw_args=args.raw_args)
    return_code = ReturnCode.Success
    for conn_name in sorted(conf.list_connection_names()):
        if not args.connections or conn_name in args.connections:
            return_code |= check_connection(conn_name, job)

    return return_code


class CheckConnectionsCommand(Command):
    description = "check all jdbc and mssql connections"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            "--connections",
            dest="connections",
            metavar="connections",
            type=str,
            nargs="*",
            help="If passed, only check the specified connections",
        )

    @staticmethod
    def do_command(args: Namespace):
        check_connections(args)



import os
import sys
from pathlib import Path
from typing import List, Optional

import dynaconf
from dynaconf.utils.boxing import DynaBox

from karadoc.common.conf.conf_box import ConfBox
from karadoc.common.exceptions import MissingConfigurationException
from karadoc.common.utils.assert_utils import assert_true

APPLICATION_NAME = "KARADOC"
APPLICATION_DESCRIPTION = "karadoc: a development environment for PySpark."
CONNECTION_GROUP = "connection"
SECRET_STORE_GROUP = "secret"  # nosec B105
REMOTE_ENV_GROUP = "remote_env"
CUSTOM_GROUP = "custom"
OBSERVABILITY_GROUP = "observability"


def get_local_data_folder_location() -> str:
    return "data"


def get_model_folder_location() -> str:
    return dynaconf.settings.get("model_dir", default="model")


def get_conf_folder_location() -> str:
    return dynaconf.settings.get("conf_dir", default="conf")


def get_libs_folder_location() -> str:
    return dynaconf.settings.get("libs", default="libs")


def get_template_package_name() -> str:
    return dynaconf.settings.get("template_package", default="model_lib.templates")


def get_warehouse_folder_location() -> str:
    return dynaconf.settings.get("warehouse_dir", default="data/hive/warehouse")


def get_analyze_timeline_dir_location() -> str:
    return dynaconf.settings.get("analyze_timeline_dir", default="data/analyze_timeline")


def get_spark_stream_dir() -> str:
    return dynaconf.settings.get("spark_stream_dir", default="data/spark/stream")


def get_streaming_checkpoint_folder_location() -> str:
    spark_stream_dir = get_spark_stream_dir()
    return f"{spark_stream_dir}/checkpoint"


def get_spark_stream_tmp_dir() -> str:
    spark_stream_dir = get_spark_stream_dir()
    return f"{spark_stream_dir}/tmp"  # nosec B108


def get_custom_command_packages() -> List[str]:
    value = dynaconf.settings.get("custom_command_packages", default=[])
    assert_true(isinstance(value, list), "Setting custom_command_packages should be a List[str]")
    return value


def get_env() -> str:
    return dynaconf.settings.current_env.lower()


def is_dev_env() -> bool:
    return get_env() == "development" or get_env() == "testing"


def enable_file_index_cache() -> bool:
    return dynaconf.settings.get("enable_file_index_cache", default=True)


def allow_missing_vars() -> bool:
    """When a job with declared variables (job.vars) is run, the values of the declared variables
    must be passed via the --vars option. This security can be disabled on an environment by setting
    allow_missing_vars = true, in which case the default values will be used when missing, instead of raising an error.

    It can be useful for development environment, but it is recommended to keep this to false in production.
    """
    return dynaconf.settings.get("allow_missing_vars", default=False)


def list_connection_names(env: Optional[str] = None) -> List[str]:
    """Return the list of all connection items for the specified environment if given,
    for the current environment otherwise."""
    connections = _get_settings_for_env(env).get(CONNECTION_GROUP)
    if connections:
        return [name for name, params in connections.items() if isinstance(params, DynaBox)]
    else:
        return []


def get_conn_conf_id(conn_name: str, group: str, env: str = None) -> str:
    """Builds the id of a connection:

    >>> get_conn_conf_id(conn_name="dummy", group=CONNECTION_GROUP, env="IC")
    'IC.connection.dummy'

    >>> get_conn_conf_id(conn_name="keyvault", group=SECRET_STORE_GROUP, env="PP")
    'PP.secret.keyvault'

    :param conn_name: Name of the connection
    :param env: environment on which to get the connection conf, use current environment if None
    :param group: configuration group
    :return:
    """
    if env is None:
        env = get_env()
    return f"{env}.{group}.{conn_name}"


def _get_settings_for_env(env: Optional[str]) -> DynaBox:
    if env is None:
        return dynaconf.settings
    else:
        return dynaconf.settings.from_env(env)


def __get_connection_conf(conn_name: str, group: str, env: Optional[str] = None) -> ConfBox:
    connections = _get_settings_for_env(env).get(group)
    if connections:
        connection_conf = connections.get(conn_name)
        if connection_conf:
            return ConfBox(connection_conf, env)

    conn_conf_id = get_conn_conf_id(conn_name, group=group, env=env)
    raise MissingConfigurationException(conf_id=conn_conf_id)


def get_connection_conf(connection_name: str, env: Optional[str] = None) -> ConfBox:
    """Return the configuration for the given connection in the specified environment.
    If no environment is specified, use the current environment.
    """
    return __get_connection_conf(connection_name, CONNECTION_GROUP, env)


def get_vault_conf(vault_name: str, env: Optional[str] = None) -> ConfBox:
    """Return the configuration for the given secret vault."""
    return __get_connection_conf(vault_name, group=SECRET_STORE_GROUP, env=env)


def get_remote_env_conf(remote_env_name: str, env: Optional[str] = None) -> ConfBox:
    """Return the configuration for the given remote environment."""
    return __get_connection_conf(remote_env_name, group=REMOTE_ENV_GROUP, env=env)


def get_custom_settings(env: Optional[str] = None) -> ConfBox:
    """Return the whole configuration object of custom configuration.
    Secret stores are supported as well.

    Example:

    In your settings.toml:

    .. code-block:: python

        [development]
        custom.my_conf_variable = "hello"
        custom.my_secret_variable.secret.keyvault = "secret-name-in-keyvault"

    In your code:

    my_variable = get_custom_conf().get("my_variable")
    my_secret_variable = get_custom_conf().get("my_variable")

    """
    custom_conf = _get_settings_for_env(env).get(CUSTOM_GROUP)
    if custom_conf is None:
        if env is None:
            env = get_env()
        raise MissingConfigurationException(conf_id=f"{env}.{CUSTOM_GROUP}")
    return ConfBox(custom_conf)


def get_action_file_name(conf_key: str) -> str:
    """Return the whole configuration object of action_file name configuration.

    Example:

    In your settings.toml:

    .. code-block:: python

        [development]
        action_file_name.spark.batch = "BATCH"

    In your code:

    my_variable = get_action_file_name("spark.batch")
    """
    action_file_name_default = {
        "spark.batch": "POPULATE",
        "spark.stream": "STREAM",
        "spark.quality_check": "QUALITY_CHECK",
        "spark.analyze_timeline": "ANALYZE",
    }
    action_file_name_conf = dynaconf.settings.get("action_file_name", default=DynaBox(action_file_name_default))
    if action_file_name_conf:
        return ConfBox(action_file_name_conf).get(conf_key)
    raise MissingConfigurationException(conf_id=f"{get_env()}.get_action_file_name")


def get_observability_conf() -> Optional[ConfBox]:
    """Return the configuration for the given remote environment."""
    settings = _get_settings_for_env(env=None)
    observability_conf = settings.get(OBSERVABILITY_GROUP)

    if observability_conf:
        return ConfBox(observability_conf)


def _add_libs_folder_to_python_path():
    libs_folder = str(Path(get_libs_folder_location()).absolute())
    sys.path.extend([libs_folder])
    if os.getenv("PYTHONPATH") is None:
        os.environ["PYTHONPATH"] = libs_folder
    else:
        os.environ["PYTHONPATH"] += os.pathsep + libs_folder



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union

from pyspark.sql import DataFrame

from karadoc.common import conf
from karadoc.common.job_core.has_spark import HasSpark, _table_name_to_hdfs_path
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.table_utils import parse_table_name
from karadoc.common.utils.assert_utils import assert_true

WASB_SCHEMES = ["wasb", "wasbs"]
ABFS_SCHEMES = ["abfs", "abfss"]


class HasInputs(HasSpark, JobBase, ABC):
    def __init__(self) -> None:
        super().__init__()
        JobBase.__init__(self)
        # Private attributes
        self.__inputs: Dict[str, Union[str, Dict[str, str]]] = {}
        """Stores the inputs as declared in the action file"""

        self.__standardized_inputs: Dict[str, Dict[str, str]] = {}
        """Stores the inputs after standardization,
        for instance { "alias": "schema_name.table_name" } is automatically
        transformed into { "alias": {"table": "schema_name.table_name"} }"""

        self.__input_warehouse_dir: str = conf.get_warehouse_folder_location()
        """Path to the folder where the data for all inputs managed by karadoc is stored"""

        self._limit_inputs: Optional[int] = None
        """When set, limits the size of every input DataFrame to this number of row as soon as it is loaded"""

        # Attributes that the user may change in action files
        self.hide_inputs: bool = False
        """Set to true to make inputs invisible in the `show_graph` command unless the `--show-hidden` option is used"""

    @property
    def inputs(self) -> Dict[str, dict]:
        return self.__inputs

    @inputs.setter
    def inputs(self, new_inputs: Dict[str, Union[str, dict]]):
        """Use this to declare your job's inputs.
        Expected format is a dictionary of (table alias, table declaration)
        The table declaration can either be it's full name ("schema_name.table.name")
        or a dictionary that can contain several entries:

        - table [Required]: full name of the table
        - format [Optional]: format of the table passed to spark
          (eg: orc, parquet, json, csv, text, or any custom format)
        - schema [Optional]: the Dataframe schema passed to spark, as a StructType
        - options [Optional]: a dictionary[str, str] of options

        Example:
        {
            "alias_1": "schema_name.table_name_1",
            "alias_2": {
                "table": "schema_name.table_name_2",
                "format": "csv",
                "schema": "id BIGINT, name STRING",
                "options": {"header": "false"},
            }
        }

        For the sake of keeping the code simpler table descriptions that are of type string are automatically
        transformed into dicts and stored in `standardized_inputs`.

        For example: { "alias": "schema_name.table_name" }
        is automatically transformed into { "alias": {"table": "schema_name.table_name"} }

        :param inputs:
        :return:
        """
        self.__inputs = new_inputs
        standardized_inputs = new_inputs.copy()
        for table_alias, input_table in new_inputs.items():
            if type(input_table) == str:
                standardized_inputs[table_alias] = {"table": input_table}
            elif type(input_table) == dict:
                if "table" not in input_table:
                    raise TypeError(f'job.input["{table_alias}"] is a dict, it should contain a "table" entry')
            else:
                raise TypeError(
                    f'job.input["{table_alias}"] should be a string or a dictionary, got {type(input_table)} instead.'
                )

        self.__standardized_inputs = standardized_inputs

    def list_input_tables(self) -> List[str]:
        """Return the list of full table names used as inputs for this job.

        :return: a list of string
        """
        return [input_table["table"] for input_table in self.__standardized_inputs.values()]

    @abstractmethod
    def get_reader(self):
        """Return a :class:`DataStreamReader` Or `DataFrameReader` that can be used to read data"""
        pass

    def read_table(self, table_alias: str, schema=None) -> DataFrame:
        """Read a table based on its alias

        :param table_alias: the alias of the table following this pattern: db_name.table_name
        :param schema: Optional, The schema used to read the data
        :return: a spark Dataframe
        """
        input_table = self.__standardized_inputs[table_alias]
        if schema is None:
            schema = input_table.get("schema")
        input_format = input_table.get("format")
        read_options = input_table.get("options")
        return self._read_input_table(input_table["table"], schema, input_format, read_options)

    def _read_input_table(
        self, table_full_name: str, schema=None, input_format: str = None, read_options: dict = None
    ) -> DataFrame:
        """Read a table based on its full name.

        End-users should NOT use this method directly.
        Instead, they MUST properly declare their `job.inputs` and use `read_table`.
        Not declaring `job.inputs` in a POPULATE would break the data lineage feature and all the beautiful stuff.
        This method can however be used in unit tests or for throw-away investigation code.

        :param table_full_name: the full name of the table
        :param schema: Optional, the schema used to read the data
        :param input_format: Optional, specify the format of the input table
                       (by default, the format declared in the input table's POPULATE is used)
        :param read_options: Optional, options passed to the DataFrameReader.
        :return: a spark Dataframe
        """
        from karadoc.common.job_core.load import load_non_runnable_action_file

        if input_format is None:
            input_job = load_non_runnable_action_file(table_full_name, type(self))
            input_format = input_job.output_format

        df = self._read_path(self.hdfs_input(table_full_name), input_format, schema, read_options)

        if self._limit_inputs is not None:
            df = df.limit(self._limit_inputs)

        return df

    def hdfs_input(self, table_name: str, remote_env_name: str = None) -> str:
        (schema_name, table_name, partition) = parse_table_name(table_name)
        if remote_env_name is not None:
            remote_input_warehouse_dir = self.__get_input_warehouse_dir_for_remote_env(remote_env_name)
        else:
            remote_input_warehouse_dir = self.__input_warehouse_dir
        return _table_name_to_hdfs_path(remote_input_warehouse_dir, schema_name, table_name, partition)

    def list_hdfs_inputs(self) -> List[str]:
        return [self.hdfs_input(table) for table in self.list_input_tables()]

    def _read_path(self, table_path, input_format, schema=None, options=None):
        assert_true(input_format is not None)

        df_reader = self.get_reader()
        if schema is not None:
            df_reader = df_reader.schema(schema)
        if options is not None:
            df_reader = df_reader.options(**options)

        df_reader = df_reader.format(input_format)
        df = df_reader.load(table_path)
        return df

    def load_inputs_as_views(self, cache=False):
        for alias, full_table_name in self.__inputs.items():
            df = self.read_table(alias)
            if cache:
                df = df.cache()
            df.createOrReplaceTempView(alias)

    def _configure_remote_input(self, remote_env_name: str):
        """Configure the job to read from a remote warehouse by modifying `job.__input_warehouse_dir`.

        :param remote_env_name:
        :return:
        """
        remote_input_warehouse_dir = self.__get_input_warehouse_dir_for_remote_env(remote_env_name)
        self.__input_warehouse_dir = remote_input_warehouse_dir

    def __get_input_warehouse_dir_for_remote_env(self, remote_env_name):
        remote_env = conf.get_remote_env_conf(remote_env_name)
        remote_env_type = remote_env["type"]
        if remote_env_type == "default":
            remote_input_warehouse_dir = remote_env["warehouse_dir"]
        elif remote_env_type in WASB_SCHEMES:
            storage_account = remote_env["storage_account"]
            storage_key = remote_env["storage_key"]
            container_name = remote_env["container_name"]
            warehouse_dir = remote_env["warehouse_dir"]
            self.spark.conf.set(
                "fs.azure.account.keyprovider.%s.blob.core.windows.net" % storage_account,
                "org.apache.hadoop.fs.azure.SimpleKeyProvider",
            )
            self.spark.conf.set("fs.azure.account.key.%s.blob.core.windows.net" % storage_account, storage_key)
            remote_input_warehouse_dir = "wasbs://%s@%s.blob.core.windows.net/%s" % (
                container_name,
                storage_account,
                warehouse_dir,
            )
        elif remote_env_type in ABFS_SCHEMES:
            storage_account = remote_env["storage_account"]
            container_name = remote_env["container_name"]
            warehouse_dir = remote_env["warehouse_dir"]
            remote_input_warehouse_dir = "abfss://%s@%s.dfs.core.windows.net/%s" % (
                container_name,
                storage_account,
                warehouse_dir,
            )
        else:
            raise NotImplementedError(
                f"env type {remote_env_type} is not supported, supported env type are {WASB_SCHEMES + ABFS_SCHEMES}"
            )
        return remote_input_warehouse_dir



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['karadoc',
 'karadoc.airflow',
 'karadoc.biqguery',
 'karadoc.cli',
 'karadoc.cli.commands',
 'karadoc.cli.commands.build',
 'karadoc.common',
 'karadoc.common.analyze',
 'karadoc.common.commands',
 'karadoc.common.commands.options',
 'karadoc.common.conf',
 'karadoc.common.exceptions',
 'karadoc.common.graph',
 'karadoc.common.job_core',
 'karadoc.common.model',
 'karadoc.common.observability',
 'karadoc.common.output',
 'karadoc.common.quality',
 'karadoc.common.run',
 'karadoc.common.stream',
 'karadoc.common.utils',
 'karadoc.common.validations',
 'karadoc.connectors',
 'karadoc.spark',
 'karadoc.test_utils']

package_data = \
{'': ['*']}

install_requires = \
['azure-keyvault>=1.1.0,<2.0.0',
 'dynaconf>=2.2.3,<3.0.0',
 'google-cloud-bigquery>=2.31.0,<3.0.0',
 'graphviz>=0.12,<0.13',
 'networkx>=2.2,<3.0',
 'tabulate>=0.8.9,<0.9.0',
 'termcolor>=1.1.0,<2.0.0',
 'xlsxwriter>=3.0.1,<4.0.0']

entry_points = \
{'console_scripts': ['karadoc = karadoc.launcher:main']}

setup_kwargs = {
    'name': 'karadoc',
    'version': '0.0.3',
    'description': 'Karadoc is a data engineering Python framework built on top of PySpark that simplifies ETL/ELT',
    'long_description': '# Karadoc\n',
    'author': 'Furcy Pin',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/FurcyPin/karadoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.11',
}


setup(**setup_kwargs)



import traceback
from argparse import ArgumentParser
from typing import Callable, Iterator, List, Optional, Type

from karadoc.common.analyze.analyze_job import AnalyzeJob
from karadoc.common.commands.command import Command
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.exceptions import ActionFileLoadingError
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.job_core.load import load_non_runnable_action_file
from karadoc.common.model import file_index
from karadoc.common.quality.quality_check_job import QualityCheckJob
from karadoc.common.run.spark_batch_job import SparkBatchJob
from karadoc.common.stream.spark_stream_job import SparkStreamJob
from karadoc.common.utils.assert_utils import assert_true
from karadoc.common.validations import ValidationResult, ValidationSeverity
from karadoc.common.validations.connection_validations import validate_connections
from karadoc.common.validations.job_validations import (
    ValidationResult_JobLoadError,
    validate_inputs,
    validate_keys,
    validate_output_partition,
)
from karadoc.common.validations.quality_check_validations import (
    validate_alert_duplicate_names,
    validate_metric_duplicate_names,
)
from karadoc.common.validations.unused_connection_validations import (
    validate_unused_connections,
)

TOTAL = "total"
OK = "ok"
FATAL = "fatal"
NON_FATAL = "non_fatal"


def inspect_results(results: List[ValidationResult]):
    """Function used for unit tests. Mock this function to be able to inspect the produced results
    during the execution of the command."""
    pass


def __get_non_fatal_count_str(non_fatal_count, severity_threshold):
    non_fatal_count_str = ""
    if non_fatal_count > 0:
        plural = "" if non_fatal_count <= 1 else "s or lower"
        non_fatal_count_str = f" {non_fatal_count} {severity_threshold.decrease().repr_item}{plural}."
    return non_fatal_count_str


def __get_fatal_count_str(fatal_count, severity_threshold):
    fatal_count_str = ""
    if fatal_count > 0:
        plural = "" if fatal_count <= 1 else "s or higher"
        fatal_count_str = f" {fatal_count} {severity_threshold.repr_item}{plural}."
    return fatal_count_str


def _validate_action_file(
    full_table_name: str, job_type: Type, validations: List[Callable], envs: Optional[List[str]]
) -> Iterator[ValidationResult]:
    """Load the action file for the specified table and job type, and apply validations on it.

    :param full_table_name: The full name of the table to validate
    :param job_type: The type of job to validate
    :param validations: The list of validation functions to apply
    :param envs: An optional list of environment against which the validations must be run
    :return:
    """
    try:
        job = load_non_runnable_action_file(full_table_name, job_type)
    except ActionFileLoadingError as e:
        yield ValidationResult_JobLoadError(error_message=e.args[0], cause=traceback.format_exc())
    else:
        for validation in validations:
            yield from validation(job, envs=envs)


def _print_results(results):
    for res in results:
        print(res.color_str() + "\n")


def _validate_action_files(
    job_type: Type, validations: List[Callable], envs: Optional[List[str]], severity_threshold=ValidationSeverity.Error
):
    assert_true(issubclass(job_type, JobBase))

    counters = {TOTAL: 0, OK: 0, FATAL: 0, NON_FATAL: 0}
    for schema, table, action_file in file_index.list_action_files(job_type):
        full_table_name = schema + "." + table
        counters[TOTAL] += 1
        results = list(_validate_action_file(full_table_name, job_type, validations, envs))

        nb_non_fatal_results = len([r for r in results if r.severity < severity_threshold])
        nb_fatal_results = len([r for r in results if r.severity >= severity_threshold])

        if len(results) > 0:
            inspect_results(results)
            print(full_table_name + " at " + action_file + " :")
            _print_results(results)

        _update_counters(counters, nb_fatal_results, nb_non_fatal_results)

    fatal_count_str = __get_fatal_count_str(counters[FATAL], severity_threshold)
    non_fatal_count_str = __get_non_fatal_count_str(counters[NON_FATAL], severity_threshold)
    print(
        f"{counters[OK]} / {counters[TOTAL]} {job_type.get_action_file_name()} "
        f"validated.{fatal_count_str}{non_fatal_count_str}"
    )

    if counters[FATAL] > 0:
        return ReturnCode.Error
    else:
        return ReturnCode.Success


def _update_counters(counters, nb_fatal_results, nb_non_fatal_results):
    if nb_fatal_results == 0 and nb_non_fatal_results == 0:
        counters[OK] += 1
    else:
        if nb_fatal_results == 0:
            counters[NON_FATAL] += 1
        else:
            counters[FATAL] += 1


def _validate_unused_connections(envs: Optional[List[str]], severity_threshold=ValidationSeverity.Error):
    results = list(validate_unused_connections(envs))
    if len(results) > 0:
        inspect_results(results)
    for res in results:
        print(res.color_str())
    non_fatal_count = len([r for r in results if r.severity < severity_threshold])
    fatal_count = len([r for r in results if r.severity >= severity_threshold])
    fatal_count_str = __get_fatal_count_str(fatal_count, severity_threshold)
    non_fatal_count_str = __get_non_fatal_count_str(non_fatal_count, severity_threshold)
    print(f"{len(results)} unused connections found.{fatal_count_str}{non_fatal_count_str}")
    if fatal_count > 0:
        return ReturnCode.Error
    else:
        return ReturnCode.Success


class ValidateCommand(Command):
    description = "apply validation checks on all tables and connections referenced in the project"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            "--coverage",
            dest="coverage",
            default=False,
            action="store_true",
            help="Activate this option to generate code coverage report on the files that were tested.",
        )
        parser.add_argument(
            "--envs",
            dest="envs",
            metavar="ENV",
            type=str,
            default=None,
            nargs="+",
            help="Validate connections against each one of the specified environments. By default, connections are only"
            "validated against the current environment.",
        )

    @staticmethod
    def do_command(args) -> ReturnCode:
        return_code = ReturnCode.Success
        cov = None
        if args.coverage:
            import coverage

            cov = coverage.Coverage()
            cov.start()

        return_code |= _validate_action_files(
            job_type=SparkBatchJob,
            validations=[
                validate_inputs,
                validate_output_partition,
                validate_connections,
                validate_keys,
            ],
            envs=args.envs,
        )
        return_code |= _validate_action_files(
            job_type=SparkStreamJob,
            validations=[
                validate_inputs,
                validate_output_partition,
                validate_connections,
            ],
            envs=args.envs,
        )
        return_code |= _validate_action_files(
            job_type=QualityCheckJob,
            validations=[
                validate_alert_duplicate_names,
                validate_metric_duplicate_names,
                validate_connections,
            ],
            envs=args.envs,
        )
        return_code |= _validate_action_files(job_type=AnalyzeJob, validations=[], envs=args.envs)

        try:
            return_code |= _validate_unused_connections(envs=args.envs)
        except ActionFileLoadingError:
            pass

        if cov is not None:
            cov.stop()
            cov.save()

        return return_code



import traceback
from typing import Generator, List, Optional, Union

from karadoc.common.conf import CONNECTION_GROUP, get_conn_conf_id
from karadoc.common.connector import (
    ConfigurableConnector,
    Connector,
    _load_connector_for_env,
)
from karadoc.common.exceptions import MissingConfigurationException
from karadoc.common.job_core.has_disable import HasDisable
from karadoc.common.job_core.has_external_inputs import HasExternalInputs
from karadoc.common.job_core.has_external_outputs import HasExternalOutputs
from karadoc.common.job_core.has_stream_external_inputs import HasStreamExternalInputs
from karadoc.common.job_core.has_stream_external_output import HasStreamExternalOutput
from karadoc.common.validations import (
    ValidationResult,
    ValidationResultTemplate,
    ValidationSeverity,
)

ValidationResult_DisabledConnection = ValidationResultTemplate(
    check_type="karadoc.connection.disabled",
    message_template="The connection {conn_conf_id} is disabled",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_MissingConnection = ValidationResultTemplate(
    check_type="karadoc.external_io.missing_connection",
    message_template='{io_type} {io_name} has no "connection" parameter',
    default_severity=ValidationSeverity.Error,
)

ValidationResult_MissingConnectionConfiguration = ValidationResultTemplate(
    check_type="karadoc.external_io.missing_connection_configuration",
    message_template="Could not find configuration {conn_conf_id}",
    default_severity=ValidationSeverity.Error,
)

ValidationResult_MissingConnectionConfigurationDisabledJob = ValidationResultTemplate(
    check_type="karadoc.external_io.missing_connection_configuration_disabled_job",
    message_template="Could not find configuration {conn_conf_id} but the job is disabled",
    default_severity=ValidationSeverity.Warning,
)

ValidationResult_ConnectionLoadError = ValidationResultTemplate(
    check_type="karadoc.connection.loading_error",
    message_template="Could not load connector for {conn_conf_id} :\n {cause}",
    default_severity=ValidationSeverity.Critical,
)


def connection_is_not_disabled(
    job: Union[HasExternalInputs, HasExternalOutputs, HasDisable], conn_conf_id: str, connection: Connector, **_
) -> Generator[ValidationResult, None, None]:
    """Yield an error if the connection is disabled and the job is not"""
    if connection.conf.get("disable") and not job.disable:
        yield ValidationResult_DisabledConnection(conn_conf_id=conn_conf_id)


def _validate_configurable_connection(connection: ConfigurableConnector, conn_conf_id: str):
    for validation_result in connection.validate_params():
        validation_result.message += f" in connection {conn_conf_id}"
        yield validation_result


def load_and_validate_external_io(
    job: HasDisable, io_type: str, io_name: str, io_conf: dict, env: Optional[str], **kwargs
) -> Generator[ValidationResult, None, None]:
    """Yield an error if the connection cannot be loaded"""
    conn_name = io_conf.get(CONNECTION_GROUP)
    conn_conf_id = get_conn_conf_id(conn_name, group=CONNECTION_GROUP, env=env)
    if conn_name is None:
        yield ValidationResult_MissingConnection(io_type=io_type, io_name=io_name)
    else:
        try:
            connection = _load_connector_for_env(conn_name, None, env)
            if isinstance(connection, ConfigurableConnector) and not connection.disable.get():
                yield from _validate_configurable_connection(connection, conn_conf_id)
        except MissingConfigurationException:
            if job.disable:
                yield ValidationResult_MissingConnectionConfigurationDisabledJob(conn_conf_id=conn_conf_id)
            else:
                yield ValidationResult_MissingConnectionConfiguration(conn_conf_id=conn_conf_id)
        except Exception:
            yield ValidationResult_ConnectionLoadError(conn_conf_id=conn_conf_id, cause=traceback.format_exc())
        else:
            yield from connection_is_not_disabled(job=job, conn_conf_id=conn_conf_id, connection=connection, **kwargs)


def __validate_connections_on_env(
    job: Union[HasExternalInputs, HasExternalOutputs, HasStreamExternalInputs, HasStreamExternalOutput, HasDisable],
    env: Optional[str],
) -> Generator[ValidationResult, None, None]:
    if isinstance(job, HasExternalInputs) or isinstance(job, HasStreamExternalInputs):
        for io_name, io_conf in job.external_inputs.items():
            yield from load_and_validate_external_io(
                job=job, io_type="external_input", io_name=io_name, io_conf=io_conf, env=env
            )
    if isinstance(job, HasExternalOutputs):
        for io_name, io_conf in job.external_outputs.items():
            yield from load_and_validate_external_io(
                job=job, io_type="external_output", io_name=io_name, io_conf=io_conf, env=env
            )
    if isinstance(job, HasStreamExternalOutput) and job.external_output is not None:
        yield from load_and_validate_external_io(
            job=job, io_type="external_output", io_name="", io_conf=job.external_output, env=env
        )


def validate_connections(
    job: Union[HasExternalInputs, HasExternalOutputs, HasStreamExternalInputs, HasStreamExternalOutput, HasDisable],
    envs: List[str] = None,
) -> Generator[ValidationResult, None, None]:
    """Performs various checks on the connections used by the given job.
     If a list of environment names is provided, the connections will be validated against each environment.
     If no list is provided, connections will only be validated against the current environment.

    :param job: A action job
    :param envs: A list of execution environments to check (set with ENV_FOR_DYNACONF)
    :return: A list of issues found
    """
    if envs is None:
        envs = [None]
    for env in envs:
        yield from __validate_connections_on_env(job, env)



import os

# When running karadoc, this setting must always be set to true
# Otherwise, nodes of the setting tree that come from other files (or that override the "default" environment)
# will erase all the other branches.
# For an example, comment this and run the unit test `test_get_custom_conf_with_dynaconf_merge`

os.environ["MERGE_ENABLED_FOR_DYNACONF"] = "true"

# For some reason, the following code sample fails when unit tests are run within PyCharm,
# but does not fail when run outside of PyCharm:
#
#    import karadoc
#    karadoc.cli.run_command(...)
#
# The import below fixes that issue.
from karadoc import cli  # noqa: E402

cli = cli

# This makes sure that the content libs/ folder is added to the python path in every situation
from karadoc.common.conf.package import _add_libs_folder_to_python_path  # noqa: E402

_add_libs_folder_to_python_path()



class HasBeforeAfter:
    def __init__(self) -> None:
        # Attributes that the user is not supposed to change
        self.__before_each = None
        self.__after_each = None
        self.__before_all = None
        self.__after_all = None

    def before_each(self) -> None:
        """This method will be called before executing each action method defined in the job's definition file.

        This is similar to unittest's `setup()` method.
        """
        if self.__before_each is not None:
            return self.__before_each()

    def after_each(self) -> None:
        """This method will be called after executing each action method defined in the job's definition file.

        This is similar to unittest's `teardown()` method.
        """
        if self.__after_each is not None:
            return self.__after_each()

    def before_all(self) -> None:
        """This method will be called once before executing all the action methods defined in the job's definition file.

        This is similar to unittest's `setupClass()` method.
        """
        if self.__before_all is not None:
            return self.__before_all()

    def after_all(self) -> None:
        """This method will be called once after executing all the action methods defined in the job's definition file.

        This is similar to unittest's `teardownClass()` method.
        """
        if self.__after_all is not None:
            return self.__after_all()



from argparse import ArgumentParser, Namespace
from itertools import groupby
from typing import List

from karadoc.common import table_utils
from karadoc.common.commands.command import Command
from karadoc.common.model import file_index
from karadoc.common.run import load_populate, print_job_properties


def list_tables():
    for (_, groups) in groupby(file_index.list_schema_table_folders(), lambda x: x[0]):
        for (schema, table, path) in groups:
            full_table_name = "%s.%s" % (schema, table)
            yield full_table_name


def list_tables_with_details():
    for full_table_name in list_tables():
        if table_utils.populate_exists(full_table_name):
            job = load_populate(full_table_name)
            print_job_properties(job)
            print()


def having_keys(full_table_name):
    if table_utils.populate_exists(full_table_name) and (
        load_populate(full_table_name).primary_key is not None or load_populate(full_table_name).secondary_keys != []
    ):
        return True
    else:
        return False


def is_disabled(full_table_name):
    return load_populate(full_table_name).disable


def having_quality_check(full_table_name):
    if table_utils.quality_check_exists(full_table_name):
        return True
    else:
        return False


def get_filter_functions(args):
    if args.with_keys:
        yield having_keys
    if args.is_disabled:
        yield is_disabled
    if args.has_quality_check:
        yield having_quality_check


def list_tables_having_indicated_details(args: List[str]):
    tables = [full_table_name for full_table_name in list_tables()]
    for fun in get_filter_functions(args):
        tables = filter(fun, tables)
    return sorted(tables)


class ListCommand(Command):
    description = "list all tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            "--with-keys", dest="with_keys", default=False, action="store_true", help="list tables with indicated keys"
        )
        parser.add_argument(
            "--is-disabled", dest="is_disabled", default=False, action="store_true", help="list disabled tables"
        )
        parser.add_argument(
            "--has-quality-check",
            dest="has_quality_check",
            default=False,
            action="store_true",
            help="list tables having quality_checks",
        )
        parser.add_argument("--output", dest="output", metavar="output", type=str, help="generate an output file")

    @staticmethod
    def do_command(args: Namespace):
        tables = list_tables_having_indicated_details(args)
        if args.output is not None:
            with open(args.output, "w") as out:
                for r in tables:
                    out.write("{}\n".format(r))
        else:
            for r in tables:
                print(r)



import difflib
from typing import Dict, List, Optional, Tuple

from pyspark.sql import Column, DataFrame, Window
from pyspark.sql import functions as f
from pyspark.sql.types import (
    ArrayType,
    IntegerType,
    LongType,
    MapType,
    StringType,
    StructField,
    StructType,
)
from pyspark.storagelevel import StorageLevel

from karadoc.common import spark_utils
from karadoc.common.spark_utils import quote, quote_columns


class NonregResult:
    def __init__(self, same_schema: bool, diff_df: Optional[DataFrame]):
        self.same_schema = same_schema
        self.diff_df = diff_df

    @property
    def same_data(self):
        return self.diff_df is None

    @property
    def is_ok(self):
        return self.same_schema and self.same_data

    def __eq__(self, other):
        if isinstance(other, NonregResult):
            return self.same_schema == other.same_schema and self.diff_df == other.diff_df
        else:
            return NotImplemented

    def __repr__(self):
        return f"NonregResult(same_schema={self.same_schema}, diff_df={self.diff_df})"


class DataframeComparator:
    DEFAULT_NB_DIFFED_ROWS = 10
    DEFAULT_MAX_STRING_LENGTH = 30
    DEFAULT_LEFT_DF_ALIAS = "left"
    DEFAULT_RIGHT_DF_ALIAS = "right"

    def __init__(
        self,
        nb_diffed_rows: int = DEFAULT_NB_DIFFED_ROWS,
        max_string_length: int = DEFAULT_MAX_STRING_LENGTH,
        left_df_alias: str = DEFAULT_LEFT_DF_ALIAS,
        right_df_alias: str = DEFAULT_RIGHT_DF_ALIAS,
    ):
        self.nb_diffed_rows = nb_diffed_rows
        self.left_df_alias = left_df_alias
        self.right_df_alias = right_df_alias
        self.max_string_length = max_string_length

    @staticmethod
    def _get_number_of_duplicates(df: DataFrame, col: str):
        """Counts the number of duplicate rows for the given column in the given DataFrame

        Example: If a column has one value present on 2 rows and another value present on 3 rows, the total number of
        duplicate is 5. If a column unique on each row, it's number of duplicates will be 0.

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> df = spark.createDataFrame(
        ...     [(1,"a"),(2,"b"),(3,"b"),(4,"c"),(5,"c"),(6,"c"),],
        ...     "`id.col` INT, `name.col` STRING")
        >>> DataframeComparator._get_number_of_duplicates(df, "id.col")
        0
        >>> DataframeComparator._get_number_of_duplicates(df, "name.col")
        5

        :param df: a Spark DataFrame
        :param col: a column name
        :return: number of duplicate rows
        """
        return (
            df.select(quote(col))
            .groupBy(quote(col))
            .count()
            .where("count > 1")
            .agg(f.coalesce(f.sum("count"), f.lit(0)).alias("nb_duplicates"))
            .take(1)[0]
            .nb_duplicates
        )

    @staticmethod
    def _get_eligible_columns_for_join(df: DataFrame) -> Dict[str, int]:
        """Identifies the column with the least duplicates, in order to use it as the id for the comparison join.

        Eligible columns are all columns of type String, Int or Bigint that have an approximate distinct count of 90%
        of the number of rows in the DataFrame. Returns None if no such column is found.

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> df = spark.createDataFrame(
        ...     [(1,"a"),(2,"b"),(3,"b"),(4,"c"),(5,"c"),(6,"c"),]
        ...     , "`id.col` INT, `name.col` STRING")
        >>> DataframeComparator._get_eligible_columns_for_join(df)
        {'id.col': 0}
        >>> df = spark.createDataFrame([(1,"a"),(1,"a")], "`id.col` INT, `name.col` STRING")
        >>> DataframeComparator._get_eligible_columns_for_join(df)
        {}

        :param df: a Spark DataFrame
        :return: The name of the columns that have more than 90% distinct values, and their corresponding number of
            of duplicates
        """
        eligible_cols = [
            col.name for col in df.schema.fields if type(col.dataType) in [StringType, IntegerType, LongType]
        ]

        distinct_count_threshold = f.lit(90.0)
        eligibility_df = df.select(
            [
                (
                    f.approx_count_distinct(quote(col)) * f.lit(100.0) / f.count(f.lit(1)) > distinct_count_threshold
                ).alias(col)
                for col in eligible_cols
            ]
        )
        columns_with_high_distinct_count = [key for key, value in eligibility_df.collect()[0].asDict().items() if value]
        cols_with_duplicates = {
            col: DataframeComparator._get_number_of_duplicates(df, col) for col in columns_with_high_distinct_count
        }
        return cols_with_duplicates

    @staticmethod
    def _merge_count_dicts(left_dict: Dict[str, int], right_dict: Dict[str, int]):
        """

        >>> DataframeComparator._merge_count_dicts({"a": 1, "b": 1}, {"a": 1, "c": 1})
        {'a': 2, 'b': 1, 'c': 1}

        :param left_dict:
        :param right_dict:
        :return:
        """
        res = left_dict.copy()
        for x in right_dict:
            if x in left_dict:
                res[x] += right_dict[x]
            else:
                res[x] = right_dict[x]
        return res

    @staticmethod
    def _automatically_infer_join_col(left_df: DataFrame, right_df: DataFrame) -> Optional[str]:
        """Identifies the column with the least duplicates, in order to use it as the id for the comparison join.

        Eligible columns are all columns of type String, Int or Bigint that have an approximate distinct count of 90%
        of the number of rows in the DataFrame. Returns None if no suche column is found.

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> left_df = spark.createDataFrame(
        ...     [(1,"a"),(2,"b"),(3,"c"),(4,"d"),(5,"e"),(6,"f")],
        ...     "`id.col` INT, `name.col` STRING")
        >>> right_df = spark.createDataFrame(
        ...     [(1,"a"),(2,"a"),(3,"b"),(4,"c"),(5,"d"),(6,"e")],
        ...     "`id.col` INT, `name.col` STRING")
        >>> DataframeComparator._automatically_infer_join_col(left_df, right_df)
        'id.col'
        >>> left_df = spark.createDataFrame([(1,"a"),(1,"a")], "`id.col` INT, `name.col` STRING")
        >>> right_df = spark.createDataFrame([(1,"a"),(1,"a")], "`id.col` INT, `name.col` STRING")
        >>> DataframeComparator._automatically_infer_join_col(left_df, right_df)


        :param left_df: a Spark DataFrame
        :param right_df: a Spark DataFrame
        :return: The name of the column with the least duplicates in both DataFrames if it has less than 10% duplicates.
        """
        left_col_dict = DataframeComparator._get_eligible_columns_for_join(left_df)
        right_col_dict = DataframeComparator._get_eligible_columns_for_join(left_df)
        merged_col_dict = DataframeComparator._merge_count_dicts(left_col_dict, right_col_dict)

        if len(merged_col_dict) > 0:
            col, nb_duplicates = sorted(merged_col_dict.items(), key=lambda x: x[1])[0]
            return col
        else:
            return None

    @staticmethod
    def _canonize_col(field: StructField, col: Column = None) -> List[Column]:
        """Applies a transformation on the specified field depending on its type.
        This ensures that the hash of the column will be constant despite it's ordering.

        :param field: the StructField object
        :param col: optional: the Column object to canonize.
        If not given we will take the column corresponding to the fields name.
        Passing this parameter is useful to refer to a column from a specific DataFrame, for instance after a join.
        :return: a list of Columns
        """
        if col is None:
            col = f.col("`" + field.name + "`")
        if type(field.dataType) == ArrayType:
            return [f.sort_array(col)]
        elif type(field.dataType) == MapType:
            return [f.sort_array(f.map_keys(col)), f.sort_array(f.map_values(col))]
        else:
            return [col]

    @staticmethod
    def _schema_to_string(
        schema: StructType, include_nullable: bool = False, include_metadata: bool = False
    ) -> List[str]:
        res = []
        for field in schema:
            s = f"{field.name} {field.dataType.simpleString().upper()}"
            if include_nullable:
                if field.nullable:
                    s += " (nullable)"
                else:
                    s += " (not nullable)"
            if include_metadata:
                s += f" {field.metadata}"
            res.append(s)
        return res

    @staticmethod
    def _compare_schemas(left_df, right_df) -> bool:
        """Compares two DataFrames schemas and print out the differences.
        Ignore the nullable and comment attributes.

        Example:

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> left_df = spark.createDataFrame([], "id INT, c1 STRING, c2 STRING, c4 ARRAY<STRUCT<a: INT, b: STRING>>")
        >>> right_df = spark.createDataFrame([], "id INT, c1 INT, c3 STRING, c4 ARRAY<STRUCT<a: INT, d: STRING>>")
        >>> res = DataframeComparator._compare_schemas(left_df, right_df)
        Schema has changed:
        @@ -1,5 +1,5 @@
        <BLANKLINE>
         id INT
        -c1 STRING
        -c2 STRING
        +c1 INT
        +c3 STRING
         c4!.a INT
        -c4!.b STRING
        +c4!.d STRING
        >>> res
        False

        :param left_df: a DataFrame
        :param right_df: another DataFrame
        :return: True if both DataFrames have the same schema, False otherwise.
        """
        left_flat_schema = spark_utils.flatten_schema(left_df.schema, explode=True)
        right_flat_schema = spark_utils.flatten_schema(right_df.schema, explode=True)

        left_string_schema = DataframeComparator._schema_to_string(left_flat_schema)
        right_string_schema = DataframeComparator._schema_to_string(right_flat_schema)

        diff = list(difflib.unified_diff(left_string_schema, right_string_schema))[2:]

        if len(diff) > 0:
            diff_string = "\n".join(diff)
            print(f"Schema has changed:\n{diff_string}")
            return False
        else:
            print(f"Schema: ok ({len(left_df.columns)} columns)")
            return True

    def _compare_rowcounts(self, left_df, right_df):
        """Compares the number of columns between two DataFrames and print out the differences

        :param left_df: a DataFrame
        :param right_df: another DataFrame
        :return: True if both DataFrames have the same number of rows, False otherwise.
        """
        left_count = left_df.count()
        right_count = right_df.count()
        diff_count = right_count - left_count
        if diff_count != 0:
            diff_string = "more" if diff_count > 0 else "less"
            same_rowcount = False
            print(
                f"Row count has changed: \n"
                f" {self.left_df_alias}: {left_count}\n"
                f" {self.right_df_alias}: {right_count} ({abs(diff_count)} {diff_string})"
            )
        else:
            same_rowcount = True
            print(f"Row count: ok ({left_count} rows)")
        return same_rowcount

    @staticmethod
    def _build_diff_dataframe(left_df: DataFrame, right_df: DataFrame, join_cols: List[str]) -> DataFrame:
        """Performs a column-by-column comparison between two DataFrames.
        The two DataFrames must have the same columns with the same ordering.
        The column `join_col` will be used to join the two DataFrames together.
        Then we build a new DataFrame with the `join_col` and for each column, a struct with three elements:
        - `left_value`: the value coming from the `left_df`
        - `right_value`: the value coming from the `right_df`
        - `same_hash`: True if both values have the same hash, False otherwise.

        Example:

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> left_df = spark.createDataFrame([
        ...         (1, 'a', 'x', 1),
        ...         (2, 'b', 'y', 2),
        ...         (3, 'c', 'z', 3)
        ...     ], "id INT, c1 STRING, c2 STRING, c3 INT"
        ... )
        >>> right_df = spark.createDataFrame([
        ...         (1, 'd', 'x', 1),
        ...         (2, 'a', 'y', 4),
        ...         (3, 'f', 'z', 3)
        ...     ], "id INT, c1 STRING, c2 STRING, c3 INT"
        ... )
        >>> left_df.show()
        +---+---+---+---+
        | id| c1| c2| c3|
        +---+---+---+---+
        |  1|  a|  x|  1|
        |  2|  b|  y|  2|
        |  3|  c|  z|  3|
        +---+---+---+---+
        <BLANKLINE>
        >>> right_df.show()
        +---+---+---+---+
        | id| c1| c2| c3|
        +---+---+---+---+
        |  1|  d|  x|  1|
        |  2|  a|  y|  4|
        |  3|  f|  z|  3|
        +---+---+---+---+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> DataframeComparator._build_diff_dataframe(left_df, right_df, ['id']).orderBy('id').show()
        +---+-------------+------------+-------------+
        | id|           c1|          c2|           c3|
        +---+-------------+------------+-------------+
        |  1|{a, d, false}|{x, x, true}| {1, 1, true}|
        |  2|{b, a, false}|{y, y, true}|{2, 4, false}|
        |  3|{c, f, false}|{z, z, true}| {3, 3, true}|
        +---+-------------+------------+-------------+
        <BLANKLINE>

        :param left_df: a DataFrame
        :param right_df: a DataFrame with the same columns
        :param join_cols: the columns to use to perform the join.
        :return: a DataFrame containing all the columns that differ, and a dictionary that gives the number of
            differing rows for each column
        """

        def get_hash(field: StructField, df: DataFrame) -> Column:
            return f.hash(*DataframeComparator._canonize_col(field, df[quote(field.name)]))

        diff = left_df.join(right_df, join_cols, "inner")

        diff_df = diff.select(
            *quote_columns(join_cols),
            *[
                f.struct(
                    left_df[quote(field.name)].alias("left_value"),
                    right_df[quote(field.name)].alias("right_value"),
                    get_hash(field, left_df).eqNullSafe(get_hash(field, right_df)).alias("same_hash"),
                ).alias(field.name)
                for field in left_df.schema.fields
                if field.name not in join_cols
            ],
        )
        # There is no need to filter rows that are fully identical because it must have been done beforehand
        return diff_df.persist(StorageLevel.DISK_ONLY)

    def _unpivot(self, diff_df: DataFrame, join_cols: List[str]):
        """Given a diff_df, builds an unpivoted version of it.
        All the values must be cast to STRING to make sure everything fits in the same column.

        >>> diff_df = __get_test_diff_df()
        >>> diff_df.show()
        +---+-------------+------------+-------------+
        | id|           c1|          c2|           c3|
        +---+-------------+------------+-------------+
        |  1|{a, d, false}|{x, x, true}| {1, 1, true}|
        |  2|{b, a, false}|{y, y, true}|{2, 4, false}|
        |  3|{c, f, false}|{z, z, true}| {3, 3, true}|
        +---+-------------+------------+-------------+
        <BLANKLINE>

        >>> from karadoc.common.nonreg import DataframeComparator
        >>> DataframeComparator()._unpivot(diff_df, join_cols=['id']).orderBy('id', 'column').show()
        +---+------+-------------+
        | id|column|         diff|
        +---+------+-------------+
        |  1|    c1|{a, d, false}|
        |  1|    c2| {x, x, true}|
        |  1|    c3| {1, 1, true}|
        |  2|    c1|{b, a, false}|
        |  2|    c2| {y, y, true}|
        |  2|    c3|{2, 4, false}|
        |  3|    c1|{c, f, false}|
        |  3|    c2| {z, z, true}|
        |  3|    c3| {3, 3, true}|
        +---+------+-------------+
        <BLANKLINE>

        :param diff_df:
        :param join_cols:
        :return:
        """

        def truncate_string(col):
            return f.when(
                f.length(col) > f.lit(self.max_string_length),
                f.concat(f.substring(col, 0, self.max_string_length - 3), f.lit("...")),
            ).otherwise(col)

        diff_df = diff_df.select(
            *quote_columns(join_cols),
            *[
                f.struct(
                    truncate_string(diff_df[quote(column) + ".left_value"].cast(StringType())).alias("left_value"),
                    truncate_string(diff_df[quote(column) + ".right_value"].cast(StringType())).alias("right_value"),
                    diff_df[quote(column) + ".same_hash"].alias("same_hash"),
                ).alias(column)
                for column in diff_df.columns
                if column not in join_cols
            ],
        )
        unpivot = spark_utils.unpivot(diff_df, pivot_columns=join_cols, key_alias="column", value_alias="diff")
        # #102035
        # We persist the unpivoted dataframe as a workaround in spark 3.1.2 because we faced OOM errors
        # in the spark optimization of the query execution plan. We can remove this workaround on spark 3.1.3
        return unpivot.persist(StorageLevel.DISK_ONLY)

    @staticmethod
    def _build_diff_per_column_df(unpivoted_diff_df: DataFrame, nb_diffed_rows) -> DataFrame:
        """Given an `unpivoted_diff_df` DataFrame, builds a DataFrame that gives for each columns the N most frequent
        differences that are happening, where N = `nb_diffed_rows`.

        Example:

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> unpivoted_diff_df = spark.createDataFrame([
        ...         (1, "c1", {'left_value': 'a', 'right_value': 'd', 'same_hash': False}),
        ...         (1, "c2", {'left_value': 'x', 'right_value': 'x', 'same_hash': True}),
        ...         (1, "c3", {'left_value': '1', 'right_value': '1', 'same_hash': True}),
        ...         (2, "c1", {'left_value': 'b', 'right_value': 'a', 'same_hash': False}),
        ...         (2, "c2", {'left_value': 'y', 'right_value': 'y', 'same_hash': True}),
        ...         (2, "c3", {'left_value': '2', 'right_value': '4', 'same_hash': False}),
        ...         (3, "c1", {'left_value': 'c', 'right_value': 'f', 'same_hash': False}),
        ...         (3, "c2", {'left_value': 'z', 'right_value': 'z', 'same_hash': True}),
        ...         (3, "c3", {'left_value': '3', 'right_value': '3', 'same_hash': True})
        ...     ], "id INT, column STRING, diff STRUCT<left_value: STRING, right_value: STRING, same_hash: BOOLEAN>"
        ... )
        >>> unpivoted_diff_df.show()
        +---+------+-------------+
        | id|column|         diff|
        +---+------+-------------+
        |  1|    c1|{a, d, false}|
        |  1|    c2| {x, x, true}|
        |  1|    c3| {1, 1, true}|
        |  2|    c1|{b, a, false}|
        |  2|    c2| {y, y, true}|
        |  2|    c3|{2, 4, false}|
        |  3|    c1|{c, f, false}|
        |  3|    c2| {z, z, true}|
        |  3|    c3| {3, 3, true}|
        +---+------+-------------+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> DataframeComparator._build_diff_per_column_df(unpivoted_diff_df, 1).orderBy('column').show()
        +------+----------+-----------+--------------+--------------------+
        |column|left_value|right_value|nb_differences|total_nb_differences|
        +------+----------+-----------+--------------+--------------------+
        |    c1|         a|          d|             1|                   3|
        |    c3|         2|          4|             1|                   1|
        +------+----------+-----------+--------------+--------------------+
        <BLANKLINE>

        :param unpivoted_diff_df: a diff DataFrame
        :return: a dict that gives for each column with differences the number of rows with different values
          for this column
        """
        # We must make sure to break ties on nb_differences when ordering to ensure a deterministic for unit tests.
        window = Window.partitionBy(f.col("column")).orderBy(
            f.col("nb_differences").desc(), f.col("left_value"), f.col("right_value")
        )

        df = (
            unpivoted_diff_df.filter("diff.same_hash = false")
            .groupBy("column", "diff.left_value", "diff.right_value")
            .agg(f.count(f.lit(1)).alias("nb_differences"))
            .withColumn("row_num", f.row_number().over(window))
            .withColumn("total_nb_differences", f.sum("nb_differences").over(Window.partitionBy("column")))
            .where(f.col("row_num") <= f.lit(nb_diffed_rows))
            .drop("row_num")
        )

        return df

    @staticmethod
    def _row_frequency_count(df: DataFrame) -> DataFrame:
        """Given a DataFrame, returns another DataFrame with the number of occurrences of each rows,
        sorted by most frequent rows.

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> df = spark.createDataFrame([
        ...         ("a", "b"),
        ...         ("a", "b"),
        ...         ("a", "b"),
        ...         ("a", "c"),
        ...         ("b", "d"),
        ...     ], "left_col: STRING, right_col: STRING"
        ... )
        >>> df.show()
        +--------+---------+
        |left_col|right_col|
        +--------+---------+
        |       a|        b|
        |       a|        b|
        |       a|        b|
        |       a|        c|
        |       b|        d|
        +--------+---------+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> DataframeComparator._row_frequency_count(df).show()
        +--------+---------+---+
        |left_col|right_col| nb|
        +--------+---------+---+
        |       a|        b|  3|
        |       a|        c|  1|
        |       b|        d|  1|
        +--------+---------+---+
        <BLANKLINE>

        :param df:
        :return:
        """
        return df.groupBy(*quote_columns(df.columns)).agg(f.count(f.lit(1)).alias("nb")).orderBy(f.desc("nb"))

    def _format_diff_df(self, join_cols: List[str], diff_df: DataFrame) -> DataFrame:
        """Given a diff DataFrame, rename the columns to prefix them with the left_df_alias and right_df_alias.

        :param join_cols:
        :param diff_df:
        :return:
        """
        return diff_df.select(
            *quote_columns(join_cols),
            *[
                col
                for col_name in diff_df.columns
                if col_name not in join_cols
                for col in [
                    diff_df[quote(col_name)]["left_value"].alias(f"{self.left_df_alias}.{col_name}"),
                    diff_df[quote(col_name)]["right_value"].alias(f"{self.right_df_alias}.{col_name}"),
                ]
            ],
        )

    def _get_diff_count_per_col(self, diff_df: DataFrame, join_cols: List[str]) -> DataFrame:
        """Given a diff_df and its list of join_cols, return a DataFrame with the following properties:

        - One row per "diff tuple" (col_name, col_value_left, col_value_right)
        - A column nb_differences that gives the number of occurrence of each "diff tuple"
        - A column total_nb_differences that gives total number of differences found for this col_name.

        >>> _diff_df = __get_test_diff_df()
        >>> _diff_df.show()
        +---+-------------+------------+-------------+
        | id|           c1|          c2|           c3|
        +---+-------------+------------+-------------+
        |  1|{a, d, false}|{x, x, true}| {1, 1, true}|
        |  2|{b, a, false}|{y, y, true}|{2, 4, false}|
        |  3|{c, f, false}|{z, z, true}| {3, 3, true}|
        +---+-------------+------------+-------------+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> df_comparator = DataframeComparator(left_df_alias="before", right_df_alias="after")
        >>> df_comparator._get_diff_count_per_col(_diff_df, join_cols = ['id']).show()
        +------+----------+-----------+--------------+--------------------+
        |column|left_value|right_value|nb_differences|total_nb_differences|
        +------+----------+-----------+--------------+--------------------+
        |    c1|         a|          d|             1|                   3|
        |    c1|         b|          a|             1|                   3|
        |    c1|         c|          f|             1|                   3|
        |    c3|         2|          4|             1|                   1|
        +------+----------+-----------+--------------+--------------------+
        <BLANKLINE>

        :param diff_df:
        :param join_cols:
        :return:
        """
        unpivoted_diff_df = self._unpivot(diff_df, join_cols)
        diff_count_per_col_df = (
            DataframeComparator._build_diff_per_column_df(unpivoted_diff_df, self.nb_diffed_rows)
            .orderBy("column")
            .cache()
        )
        return diff_count_per_col_df

    def _display_diff_results(self, diff_count_per_col_df: DataFrame, join_cols: List[str]):
        """Displays the results of the diff analysis.
        We first display the a summary of all columns that changed with the number of changes,
        then for each column, we display a summary of the most frequent changes and then
        we display examples of rows where this column changed, along with all the other columns
        that changed in this diff.

        Example:

        >>> spark = spark_utils.get_spark_session("doctest")
        >>> diff_count_per_col_df = spark.createDataFrame([
        ...         ('c1', 'a', 'd', 1, 3),
        ...         ('c3', '2', '4', 1, 1),
        ...     ], "column STRING, left_value STRING, right_value STRING, nb_differences INT, total_nb_differences INT"
        ... )
        >>> diff_count_per_col_df.show()
        +------+----------+-----------+--------------+--------------------+
        |column|left_value|right_value|nb_differences|total_nb_differences|
        +------+----------+-----------+--------------+--------------------+
        |    c1|         a|          d|             1|                   3|
        |    c3|         2|          4|             1|                   1|
        +------+----------+-----------+--------------+--------------------+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> df_comparator = DataframeComparator(left_df_alias="before", right_df_alias="after")
        >>> df_comparator._display_diff_results(diff_count_per_col_df, join_cols = ['id'])
        Found the following differences by joining with column 'id'
        +-------------+---------------+------+-----+----------------+
        |`column name`|`total nb diff`|before|after|`nb differences`|
        +-------------+---------------+------+-----+----------------+
        |c1           |3              |a     |d    |1               |
        |c3           |1              |2     |4    |1               |
        +-------------+---------------+------+-----+----------------+
        <BLANKLINE>

        :param diff_count_per_col_df:
        :param join_cols:
        :return:
        """
        if len(join_cols) == 1:
            print(f"Found the following differences by joining with column '{join_cols[0]}'")
        else:
            print(f"Found the following differences by joining with columns: {join_cols}")
        diff_count_per_col_df.select(
            f.col("column").alias("`column name`"),
            f.col("total_nb_differences").alias("`total nb diff`"),
            f.col("left_value").alias(self.left_df_alias),
            f.col("right_value").alias(self.right_df_alias),
            f.col("nb_differences").alias("`nb differences`"),
        ).show(1000 * self.nb_diffed_rows, False)

    def _display_diff_examples(self, diff_df: DataFrame, diff_count_per_col_df: DataFrame, join_cols: List[str]):
        """For each column that has differences, print examples of rows where such a difference occurs.

        >>> _diff_df = __get_test_diff_df()
        >>> _diff_df.show()
        +---+-------------+------------+-------------+
        | id|           c1|          c2|           c3|
        +---+-------------+------------+-------------+
        |  1|{a, d, false}|{x, x, true}| {1, 1, true}|
        |  2|{b, a, false}|{y, y, true}|{2, 4, false}|
        |  3|{c, f, false}|{z, z, true}| {3, 3, true}|
        +---+-------------+------------+-------------+
        <BLANKLINE>
        >>> from karadoc.common.nonreg import DataframeComparator
        >>> df_comparator = DataframeComparator(left_df_alias="before", right_df_alias="after")
        >>> _diff_count_per_col_df = df_comparator._get_diff_count_per_col(_diff_df, join_cols = ['id'])
        >>> df_comparator._display_diff_examples(_diff_df, _diff_count_per_col_df, join_cols = ['id'])
        Detailed examples :
        'c1' : 3 rows
        +---+---------+--------+---------+--------+
        |id |before.c1|after.c1|before.c3|after.c3|
        +---+---------+--------+---------+--------+
        |1  |a        |d       |1        |1       |
        |2  |b        |a       |2        |4       |
        |3  |c        |f       |3        |3       |
        +---+---------+--------+---------+--------+
        <BLANKLINE>
        'c3' : 1 rows
        +---+---------+--------+---------+--------+
        |id |before.c1|after.c1|before.c3|after.c3|
        +---+---------+--------+---------+--------+
        |2  |b        |a       |2        |4       |
        +---+---------+--------+---------+--------+
        <BLANKLINE>

        :param diff_df:
        :param diff_count_per_col_df:
        :param join_cols:
        :return:
        """
        rows = diff_count_per_col_df.select("column", "total_nb_differences").distinct().collect()
        diff_count_per_col = [(r[0], r[1]) for r in rows]
        print("Detailed examples :")
        for col, nb in diff_count_per_col:
            print(f"'{col}' : {nb} rows")
            rows_that_changed_for_that_column = diff_df.filter(~diff_df[quote(col)]["same_hash"]).select(
                *join_cols, *[quote(r[0]) for r in rows]
            )
            self._format_diff_df(join_cols, rows_that_changed_for_that_column).show(self.nb_diffed_rows, False)

    def _analyze_differences(
        self, left_only: DataFrame, right_only: DataFrame, join_cols: List[str]
    ) -> Tuple[Optional[DataFrame], Optional[List[str]]]:
        """Performs an in-depth analysis between two DataFrames with the same columns and prints the differences found.
        We first attempt to identify columns that look like ids.
        For that we choose all the columns with an approximate_count_distinct greater than 90% of the row count.
        For each column selected this way, we then perform a join and compare the DataFrames column by column.

        :param left_only: a DataFrame
        :param right_only: another DataFrame with the same columns
        :param join_cols: the list of columns on which to perform the join
        :return: a Dict that gives for each eligible join column the corresponding diff DataFrame
        """
        if join_cols is None:
            inferred_join_col = DataframeComparator._automatically_infer_join_col(left_only, right_only)
            if inferred_join_col is not None:
                print(
                    f"We will try to find the differences by joining the DataFrames "
                    f"using this column: {inferred_join_col}"
                )
                return self._build_diff_dataframe(left_only, right_only, [inferred_join_col]), [inferred_join_col]
            else:
                print(
                    "Could not automatically infer a column for joining the DataFrames.\n"
                    "Please provide one by using the --nonreg-join-cols options or by setting the job.primary_key "
                    "in the table's POPULATE.py file."
                )
                return None, None
        else:
            print(
                f"We will try to find the differences by joining the DataFrames "
                f"using the provided columns together : {join_cols}"
            )
            return self._build_diff_dataframe(left_only, right_only, join_cols), join_cols

    def compare_df(
        self, left_df: DataFrame, right_df: DataFrame, join_cols: List[str] = None, show_examples: bool = False
    ) -> NonregResult:
        """Compares two DataFrames and print out the differences.
        We first compare the DataFrame schemas and row counts,
        then we compute the hashes of every rows in both DataFrames
        and compare the differences.
        If differences are found, a more in-depth analysis is performed
        to try to find which columns differ between the two DataFrames.

        If `join_cols` is specified, we will use the specified columns to perform the comparison join.
        If not, the columns to use for the join will be automatically inferred. If several columns are eligible for
        the join, the join will be performed on each of them. The automatic inference can only determine join based
        on a single column. If the DataFrame's unique keys are composite (multiple columns) they have to be given
        explicitely via `join_cols` to perform the diff analysis.

        Known Limitations :
        This method is able to compare Arrays and Maps while ignoring their ordering,
        It is able to handle nested non-repeated structures, such as STRUCT<STRUCT<>>, or even ARRAY<STRUCT>>
        but it doesn't support nested repeated structures, such as ARRAY<ARRAY<>>.

        :param left_df: a DataFrame
        :param right_df: another DataFrame
        :param join_cols: [Optional] specifies the columns on which the two DataFrames should be joined to compare them
        :param show_examples: if set to True, print for each column examples of full rows where this column changes
        :return: True if both DataFrames matches, False otherwise.
        """
        diff_df = None

        if join_cols == []:
            join_cols = None

        magic_hash_col_name = "MAGIC_HASH"

        same_schema = self._compare_schemas(left_df, right_df)

        if join_cols is not None:
            left_df = left_df.repartition(*join_cols)
        if join_cols is not None:
            right_df = right_df.repartition(*join_cols)

        left_flat = spark_utils.flatten(left_df)
        right_flat = spark_utils.flatten(right_df)

        if not same_schema:
            common_cols = spark_utils.get_common_columns(left_flat.schema, right_flat.schema)
            left_flat = left_flat.select(common_cols)
            right_flat = right_flat.select(common_cols)

        left_flat = left_flat.withColumn(
            magic_hash_col_name,
            f.hash(*[col for field in left_flat.schema.fields for col in self._canonize_col(field)]),
        )
        right_flat = right_flat.withColumn(
            magic_hash_col_name,
            f.hash(*[col for field in right_flat.schema.fields for col in self._canonize_col(field)]),
        )

        left_flat = left_flat.persist(StorageLevel.DISK_ONLY)
        right_flat = right_flat.persist(StorageLevel.DISK_ONLY)

        self._compare_rowcounts(left_flat, right_flat)

        left_hash = left_flat.select(magic_hash_col_name)
        right_hash = right_flat.select(magic_hash_col_name)

        left_hash_only = left_hash.join(right_hash, magic_hash_col_name, "leftanti")
        right_hash_only = right_hash.join(left_hash, magic_hash_col_name, "leftanti")

        left_hash_only_nb = left_hash_only.count()
        right_hash_only_nb = right_hash_only.count()

        if left_hash_only_nb > 0 or right_hash_only_nb > 0:
            print("Diff: not ok")
            if right_hash_only_nb > 0:
                print(f"{right_hash_only_nb} rows appeared")
            if left_hash_only_nb > 0:
                print(f"{left_hash_only_nb} rows disappeared")

            print("analyzing differences...")
            # We only compare the rows of the DataFrames with a hashcode that doesn't match the other DataFrame
            left_only_df = left_flat.join(left_hash_only, magic_hash_col_name, "leftsemi").drop(magic_hash_col_name)
            right_only_df = right_flat.join(right_hash_only, magic_hash_col_name, "leftsemi").drop(magic_hash_col_name)
            left_only_df = left_only_df.persist(StorageLevel.DISK_ONLY)
            right_only_df = right_only_df.persist(StorageLevel.DISK_ONLY)

            diff_df, join_cols = self._analyze_differences(left_only_df, right_only_df, join_cols)
            if diff_df is not None and diff_df.count() > 0:
                diff_count_per_col_df = self._get_diff_count_per_col(diff_df, join_cols)
                self._display_diff_results(diff_count_per_col_df, join_cols)
                if show_examples:
                    self._display_diff_examples(diff_df, diff_count_per_col_df, join_cols)

            left_only_df.unpersist()
            right_only_df.unpersist()
        else:
            print("Diff: ok")

        left_df.unpersist()
        right_df.unpersist()
        left_flat.unpersist()
        right_flat.unpersist()

        return NonregResult(same_schema, diff_df)


def __get_test_diff_df() -> DataFrame:
    spark = spark_utils.get_spark_session("doctest")
    diff_df = spark.createDataFrame(
        [
            (
                1,
                {"left_value": "a", "right_value": "d", "same_hash": False},
                {"left_value": "x", "right_value": "x", "same_hash": True},
                {"left_value": 1, "right_value": 1, "same_hash": True},
            ),
            (
                2,
                {"left_value": "b", "right_value": "a", "same_hash": False},
                {"left_value": "y", "right_value": "y", "same_hash": True},
                {"left_value": 2, "right_value": 4, "same_hash": False},
            ),
            (
                3,
                {"left_value": "c", "right_value": "f", "same_hash": False},
                {"left_value": "z", "right_value": "z", "same_hash": True},
                {"left_value": 3, "right_value": 3, "same_hash": True},
            ),
        ],
        "id INT,"
        + "c1 STRUCT<left_value: STRING, right_value: STRING, same_hash: BOOLEAN>,"
        + "c2 STRUCT<left_value: STRING, right_value: STRING, same_hash: BOOLEAN>,"
        + "c3 STRUCT<left_value: INT, right_value: INT, same_hash: BOOLEAN>",
    )
    return diff_df



import argparse
from typing import Sequence


class StoreDictKeyValue(argparse.Action):
    """Custom action for argparse that will parse a list of key=value string and build a dict out of it.

    Example:

    >>> import argparse
    >>> parser = argparse.ArgumentParser()
    >>> parser.add_argument(
    ...     '--vars',
    ...     nargs='*',
    ...     dest='vars',
    ...     metavar='KEY=VALUE',
    ...     required=False,
    ...     action=StoreDictKeyValue,
    ...     help='Set a variable value. Can be repeated to define multiple variables.'
    ... ) # doctest:+ELLIPSIS
    StoreDictKeyValue(...)

    (ELLIPSIS tells doctest to match "..." with anything)

    >>> args = parser.parse_args(["--vars", "A=1"])
    >>> args.vars
    {'A': '1'}
    >>> args = parser.parse_args(["--vars", "B=2"])
    >>> args.vars
    {'B': '2'}

    """

    def __init__(self, option_strings: Sequence[str], dest: str, **kwargs) -> None:
        super().__init__(option_strings, dest, **kwargs)
        if self.default is not None:
            raise ValueError("No default value is allowed for StoreDictKeyValue action")
        self.default = dict()

    def __call__(self, parser, namespace, values, option_string=None) -> None:
        dest = dict()
        for value in values:
            k, v = value.split("=")
            dest[k] = v
        setattr(namespace, self.dest, dest)



from typing import List, Optional, Tuple, Union


class HasKeys:
    def __init__(self) -> None:
        # Attributes that the user may change
        self.__primary_key: Optional[Union[Tuple[str, ...], str]] = None
        """Primary_key is a column or group of columns that must be unique and non-null.
        It is automatically used by the nonreg to perform joins."""
        self.__standardized_primary_key: Optional[Union[Tuple[str, ...]]] = None

        self.__secondary_keys: List[Union[Tuple[str, ...], str]] = []
        """Secondary_keys is a list of additional columns or group of columns that must be unique, non-null"""
        self.__standardized_secondary_keys: List[Tuple[str, ...]] = []

    @staticmethod
    def __calculate_standardized_key(new_key: Union[Tuple[str, ...], str]) -> Tuple[str, ...]:
        if isinstance(new_key, str):
            return (new_key,)
        elif isinstance(new_key, tuple):
            return new_key
        else:
            raise TypeError("Key type must be string or tuple[str,...]")

    @property
    def primary_key(self) -> Optional[Union[Tuple[str, ...], str]]:
        """Primary key for the job: must be non-null and unique.

        It is recommended to use 'job.standardized_primary_key' rather than this getter, unless if it is to get
        the raw primary_key definition set by the user.

        For the sake of keeping the code simpler, standardized_primary_key is defined automatically from primary_key.
        It is a transformed representation of primary_key used as parameters for some functions.
        """
        return self.__primary_key

    @primary_key.setter
    def primary_key(self, new_primary_key: Union[Tuple[str, ...], str]) -> None:
        """Use this to declare the job's primary_key.

        Expected format is a string or a tuple of strings for composite keys
        Examples:
            - job.primary_key = "column_name_1"
            - job.primary_key = ("column_name_1", "column_name_2")

        For the sake of keeping the code simpler, standardized_primary_key is defined
        automatically from primary_key. It is a transformed representation of primary_key used as parameters
        for some functions.

        """
        self.__primary_key = new_primary_key
        self.__standardized_primary_key = self.__calculate_standardized_key(new_primary_key)

    @property
    def standardized_primary_key(self) -> Optional[Tuple[str, ...]]:
        """Primary key for the job in a standardized format.

        Examples:
        >>> from karadoc.common import Job
        >>> job = Job()
        >>> job.primary_key = 'column_name_1'
        >>> job.standardized_primary_key
        ('column_name_1',)

        >>> job.primary_key = ('column_name_1', 'column_name_2')
        >>> job.standardized_primary_key
        ('column_name_1', 'column_name_2')
        """
        return self.__standardized_primary_key

    @property
    def secondary_keys(self) -> List[Union[Tuple[str, ...], str]]:
        """Secondary keys for the job: must be non-null and unique.

        It is recommended to use 'job.standardized_secondary_keys' rather than this getter, unless if it is to get
        the raw secondary_keys definition set by the user.

        For the sake of keeping the code simpler, standardized_secondary_keys is defined automatically
        from secondary_keys. It is a transformed representation of secondary_keys used as parameters for some functions.
        """
        return self.__secondary_keys

    @secondary_keys.setter
    def secondary_keys(self, new_secondary_keys: List[Union[Tuple[str, ...], str]]) -> None:
        """Use this to declare the job's secondary_keys.
        Expected format is List(Union[Tuple[str, ...], str])

        Examples:
            - secondary_keys = ["column_name_1", "column_name_2" , "column_name_3"]
            - secondary_keys = [("column_name_1", "column_name_2") , "column_name_3" ]
            - secondary_keys = [("column_name_1", "column_name_2") , ("column_name_3", "column_name_4")]

        For the sake of keeping the code simpler, standardized_secondary_keys are defined
        automatically from secondary_keys. They are a transformed representation of secondary_keys used as parameters
        for functions.

        Examples:
            if secondary_keys = [("column_name_1", "column_name_2") , "column_name_2" ]
            then standardized_secondary_keys = [("column_name_1", "column_name_2") , ("column_name_2",)]

        """
        self.__secondary_keys = new_secondary_keys
        self.__standardized_secondary_keys = []
        for key in new_secondary_keys:
            self.__standardized_secondary_keys.append(self.__calculate_standardized_key(key))

    @property
    def standardized_secondary_keys(self) -> List[Tuple[str, ...]]:
        return self.__standardized_secondary_keys



import importlib.machinery
import inspect
import sys
from os.path import isfile, splitext
from pathlib import Path
from types import FunctionType, ModuleType
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union, cast

from karadoc.common import conf
from karadoc.common.exceptions import ActionFileLoadingError, ForbiddenActionError
from karadoc.common.job_core.has_before_after import HasBeforeAfter
from karadoc.common.job_core.has_external_inputs import (
    HasExternalInputs,
    _read_external_input_signature_check,
    _read_external_inputs_signature_check,
)
from karadoc.common.job_core.has_external_outputs import (
    HasExternalOutputs,
    _write_external_output_signature_check,
    _write_external_outputs_signature_check,
)
from karadoc.common.job_core.has_spark import HasSpark
from karadoc.common.job_core.has_stream_external_inputs import (
    HasStreamExternalInputs,
    _read_stream_external_input_signature_check,
    _read_stream_external_inputs_signature_check,
)
from karadoc.common.job_core.has_stream_external_output import (
    HasStreamExternalOutput,
    _write_stream_external_output_signature_check,
)
from karadoc.common.job_core.has_vars import HasVars
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.model import file_index
from karadoc.common.quality.checks import Alert, Metric
from karadoc.common.table_utils import parse_table_name
from karadoc.common.utils.assert_utils import assert_true

Job = TypeVar("Job", JobBase, JobBase)
A = TypeVar("A")
B = TypeVar("B")


def load_non_runnable_action_file(full_table_name: str, job_type: Type[Job]) -> Job:
    """Return a non-runnable version of the `job` object declared in the ACTION_FILE.py file of the given table
    with its default variables.

    This method is useful to load any ACTION_FILE for inspecting its metadata (input, output, output_partitions, etc.)
    even when no variable is defined, by using its default values. The `run/stream/analyze` method is removed to prevent
    any misuse where one would call the ACTION_FILE without setting the variables properly.

    If you want to execute the `run/stream/analyze` method use the `load_runnable_action_file` method instead.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :param job_type: job type defined in the action file. (e.g. SparkBathJob, QualityCheckJob ...)
    :return: the SparkBatchJob object defined in the ACTION_FILE.
    """
    passed_vars = None

    job = __load_action_file(job_type, full_table_name, passed_vars)
    if job._run_method_name is not None:

        def empty_run() -> None:
            raise ForbiddenActionError(
                f"The {job._run_method_name} method of a job returned by the `load_non_runnable_action_file` "
                "method cannot be called. use load_runnable_action_file instead."
            )

        job.__setattr__(job._run_method_name, empty_run)
    return job


def load_runnable_action_file(full_table_name: str, job_type: Type[Job], passed_vars: Dict[str, str]) -> Job:
    """Return the `job` object declared in the ACTION_FILE.py file of the given table and override its variables,
    checking that they match the variables declared in the ACTION_FILE.py file.

    Explanation:

    When you call a command with the '--vars' option, this method will be called and the `vars` passed to
     the ACTION_FILE file in the job.vars dict.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :param job_type: job type defined in the action file. (e.g. SparkBathJob, QualityCheckJob ...)
    :param passed_vars: Set the variables of the job and check that they match the ones declared
    :return: the ActionFileJob object defined in the ACTION_FILE.py file
    """
    if passed_vars is None:
        passed_vars = dict()
    return __load_action_file(job_type, full_table_name, passed_vars)


def __load_action_file(job_type: Type[Job], full_table_name: str, passed_vars: Optional[Dict[str, str]]) -> Job:
    assert_true(issubclass(job_type, JobBase))
    (schema_name, table_name, _) = parse_table_name(full_table_name)
    file_path = file_index.get_action_file(schema_name, table_name, job_type)
    if file_path is None:
        expected_path = Path(f"**/{schema_name}.db") / table_name / f"{job_type.get_action_file_name()}.py"
        raise FileNotFoundError(
            f"Found no file matching the path '{expected_path}' for the table {full_table_name} "
            f"in the directory '{conf.get_model_folder_location()}'"
        )
    file_path = cast(str, file_path)
    job = __load_file(schema_name, table_name, file_path, job_type.get_action_file_name(), passed_vars, job_type)
    return job


def check_method_signatures(method_name: str, actual: FunctionType, expected: FunctionType) -> None:
    """Ensures that the signature of a given method matches the expected signature taken from another method.

    :param method_name: name of the method to check
    :param actual: function to check
    :param expected: control function with the expected signature
    :return:
    """
    if not isinstance(actual, FunctionType):
        raise TypeError("%s is not a function" % method_name)
    actual_args = inspect.getfullargspec(actual)
    expected_args = inspect.getfullargspec(expected)
    if actual_args != expected_args:
        expected_signature = (
            inspect.getsource(expected).split("\n")[0].replace("def _", "def ").replace("_signature_check", "")
        )
        raise TypeError(
            'The method {name} should have the following signature\n"{signature}"'.format(
                name=method_name, signature=expected_signature
            )
        )


def __set_job_attr_if_exists(mod: ModuleType, attr_name: str) -> None:
    if hasattr(mod, attr_name):
        setattr(mod.job, attr_name, getattr(mod, attr_name))


def __delete_module_if_exists(module_name: str) -> None:
    if module_name in sys.modules:
        del sys.modules[module_name]


def __delete_template_modules() -> None:
    """
    Delete all template modules to avoid caching templates.
    Doing so will ensure the execution of the template content every load.
    """
    from karadoc.common import conf

    template_package_name = conf.get_template_package_name()
    template_modules = [m for m in sys.modules if m.startswith(template_package_name)]
    for m in template_modules:
        del sys.modules[m]


def __load_module_file(module_name: str, module_path: Union[str, Path]) -> ModuleType:
    """Loads an action file.

    Implementation details
    ----------------------

    If an __init__.py file is located in the same folder, we load the __init__.py file first.
    This allows relative imports inside action files.

    If a module with the same name was already loaded, we must unload it first.
    This does not happen in regular uses, but it may happen while running unit test suites

    :param module_name: Name of the module that will be loaded. It must be unique. Dots will be replaced by colons.
    :param module_path: Path of the action file to load
    :return: a Python module
    """
    # The module_name must not contain dots, this tends to cause errors in relative imports
    module_name = module_name.replace(".", ":")
    if isinstance(module_path, str):
        module_path = Path(module_path)
    file_name = splitext(module_path.name)[0]
    init_file = module_path.parent / "__init__.py"
    if isfile(init_file):
        __delete_module_if_exists(module_name)
        importlib.machinery.SourceFileLoader(module_name, str(init_file)).load_module()
    submodule_name = module_name + "." + file_name
    __delete_module_if_exists(submodule_name)
    mod = importlib.machinery.SourceFileLoader(submodule_name, str(module_path)).load_module()
    return mod


def __load_optional_method(
    mod: ModuleType, tpe: type, method_name: str, signature_method: Optional[Callable[..., Any]] = None
) -> None:
    if hasattr(mod, method_name):
        private_method_name = "_" + tpe.__name__ + "__" + method_name
        if signature_method is not None:
            signature_method = cast(FunctionType, signature_method)
            check_method_signatures(method_name, getattr(mod, method_name), signature_method)
        setattr(mod.job, private_method_name, getattr(mod, method_name))


def _set_spark_batch_job(mod: ModuleType) -> None:
    from karadoc.common.run.spark_batch_job import SparkBatchJob

    if isinstance(mod.job, SparkBatchJob):
        mod.job.run = mod.run


def _set_spark_stream_job(mod: ModuleType) -> None:
    from karadoc.common.stream.spark_stream_job import SparkStreamJob

    if isinstance(mod.job, SparkStreamJob):
        mod.job.stream = mod.stream


def _set_analyze_job(mod: ModuleType) -> None:
    from karadoc.common.analyze.analyze_job import AnalyzeJob

    if isinstance(mod.job, AnalyzeJob):
        mod.job.analyze = mod.analyze


def _set_quality_check_job(mod: ModuleType) -> None:
    from karadoc.common.quality.quality_check_job import QualityCheckJob

    if isinstance(mod.job, QualityCheckJob):
        __load_optional_method(mod, HasBeforeAfter, "before_all")
        __load_optional_method(mod, HasBeforeAfter, "after_all")
        __load_optional_method(mod, HasBeforeAfter, "before_each")
        __load_optional_method(mod, HasBeforeAfter, "after_each")
        found_alerts = [obj for name, obj in inspect.getmembers(mod) if isinstance(obj, Alert)]
        for alert in found_alerts:
            mod.job.add_alert(alert)
        found_metrics = [obj for name, obj in inspect.getmembers(mod) if isinstance(obj, Metric)]
        for metric in found_metrics:
            mod.job.add_metric(metric)


def _set_has_external_inputs(mod: ModuleType) -> None:
    if isinstance(mod.job, HasExternalInputs):
        __load_optional_method(mod, HasExternalInputs, "read_external_input", _read_external_input_signature_check)
        __load_optional_method(mod, HasExternalInputs, "read_external_inputs", _read_external_inputs_signature_check)


def _set_has_stream_external_inputs(mod: ModuleType) -> None:
    if isinstance(mod.job, HasStreamExternalInputs):
        __load_optional_method(
            mod, HasStreamExternalInputs, "read_external_input", _read_stream_external_input_signature_check
        )
        __load_optional_method(
            mod, HasStreamExternalInputs, "read_external_inputs", _read_stream_external_inputs_signature_check
        )


def _set_has_external_outputs(mod: ModuleType) -> None:
    if isinstance(mod.job, HasExternalOutputs):
        __load_optional_method(mod, HasExternalOutputs, "write_external_output", _write_external_output_signature_check)
        __load_optional_method(
            mod, HasExternalOutputs, "write_external_outputs", _write_external_outputs_signature_check
        )


def _set_has_stream_external_output(mod: ModuleType) -> None:
    if isinstance(mod.job, HasStreamExternalOutput):
        __load_optional_method(
            mod, HasStreamExternalOutput, "write_external_output", _write_stream_external_output_signature_check
        )


def __set_job_from_module(mod: ModuleType) -> JobBase:
    _set_spark_batch_job(mod)
    _set_spark_stream_job(mod)
    _set_analyze_job(mod)
    _set_quality_check_job(mod)
    _set_has_external_inputs(mod)
    _set_has_stream_external_inputs(mod)
    _set_has_external_outputs(mod)
    _set_has_stream_external_output(mod)
    return mod.job


def __load_job_module(module_name: str, module_path: Union[str, Path]) -> JobBase:
    __delete_template_modules()
    mod = __load_module_file(module_name, module_path)
    job = __set_job_from_module(mod)
    return job


def __load_file(
    schema_name: str,
    table_name: str,
    file_path: str,
    file_type: str,
    passed_vars: Optional[Dict[str, str]],
    job_type: type,
) -> JobBase:
    full_table_name = schema_name + "." + table_name
    if file_path is None:
        raise ActionFileLoadingError(f"Could not find a {file_type} for table {full_table_name}")
    try:
        HasVars._global_vars = passed_vars
        HasSpark._global_init_lock = True
        job = __load_job_module(conf.APPLICATION_NAME.upper() + ":" + full_table_name, file_path)
        job.output = full_table_name
    except Exception as e:
        raise ActionFileLoadingError(
            f"Could not load {file_type} file for table {full_table_name} at {file_path}"
        ) from e
    finally:
        HasSpark._global_init_lock = False
        HasVars._global_vars = None
    if not isinstance(job, job_type):
        raise ActionFileLoadingError(
            f"The 'job' object returned by this {file_type} is not of the expected type."
            f"Expected Type: {job_type}, Type Found: {type(job)}"
        )
    return job



import logging
import sys  # noqa: F401 (used only in doctest)
from logging import Formatter, StreamHandler  # noqa: F401 (used only in doctest)
from typing import Optional


def add_logging_level(level_name: str, level_num: int, method_name: Optional[str] = None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> LOG = logging.getLogger(__name__)
    >>> LOG.setLevel('TRACE')
    >>> stdout_handler = StreamHandler(stream=sys.stdout)
    >>> stdout_handler.setFormatter(Formatter(fmt="%(message)s"))
    >>> LOG.addHandler(stdout_handler)
    >>> LOG.propagate = False # This prevents the message from being logged by handlers added to the root logger.
    >>> LOG.trace('this works')
    this works

    """
    # This method was taken from this Stack Overflow answer:
    # https://stackoverflow.com/q/2183233/35804945#35804945
    if hasattr(logging, level_name):
        return
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError("{} already defined in logging module".format(level_name))
    if hasattr(logging, method_name):
        raise AttributeError("{} already defined in logging module".format(method_name))
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError("{} already defined in logger class".format(method_name))

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)



from typing import Dict, Optional

from pyspark.sql import DataFrame
from pyspark.sql.streaming import DataStreamWriter

from karadoc.common import conf
from karadoc.common.job_core.has_spark import _table_name_to_hdfs_path
from karadoc.common.job_core.load import (
    load_non_runnable_action_file,
    load_runnable_action_file,
)
from karadoc.common.output.printing import pretty_format
from karadoc.common.stream.spark_stream_job import SparkStreamJob
from karadoc.common.table_utils import parse_table_name


def load_stream_file(full_table_name: str) -> SparkStreamJob:
    """Return a non-runnable version of the `job` object declared in the STREAM.py file of the given table
    with its default variables.

    This method is useful to load any STREAM for inspecting its metadata (input, output, output_partitions, etc.)
    even when no variable is defined, by using its default values. The `stream` method is removed to prevent any misuse
    where one would call the STREAM without setting the variables properly.

    If you want to execute the `stream` method use the `load_runnable_stream_file` method instead.

    :param full_table_name: Full name of the table (schema_name.table_name)
    :return: the SparkBatchJob object defined in the populate.
    """
    return load_non_runnable_action_file(full_table_name, SparkStreamJob)


def load_runnable_stream_file(full_table_name: str, passed_vars: Optional[Dict[str, str]]) -> SparkStreamJob:
    """Return the `job` object declared in the STREAM.py file of the given table, override its variables
    and check the load variables match those declared in the
    :param full_table_name: Full name of the table (schema_name.table_name)
    :param passed_vars: Set the variables of the job and check that they match the ones declared
    :return: the SparkStreamJob object defined in the STREAM.py file
    """
    return load_runnable_action_file(full_table_name, SparkStreamJob, passed_vars)


def print_job_properties(job: SparkStreamJob):
    if job.external_inputs != {}:
        print("external_inputs :")
        print(pretty_format(job.external_inputs))
    if job.external_output is not None:
        print("external_outputs:")
        print(pretty_format(job.external_output))


def get_checkpoint_location(output: str):
    (schema_name, table_name, _) = parse_table_name(output)
    checkpoint_folder_location = conf.get_streaming_checkpoint_folder_location()
    return _table_name_to_hdfs_path(checkpoint_folder_location, schema_name, table_name, None)


def write_stream_output(job: SparkStreamJob, df, no_export: bool, streaming_mode: str, batch_interval: Optional[str]):
    """Serialize the output of a stream if it exists and is not empty, and if its schema is correctly defined.
    By default, the output is written as orc, but future development might allow the user to specify parquet instead.
    """
    if df and type(df) is DataFrame:
        if df.columns:
            if no_export:
                print("WARN: The job export has been disabled.")
            if job.external_output is not None and not no_export:
                ds_writer = job.write_external_output(df, job.external_output)
            else:
                ds_writer = job.write_table(df)
            __start_stream_writer(ds_writer, job, streaming_mode, batch_interval)
        else:
            print("WARN: The job returned a DataFrame with zero columns. The output will not be serialized.")


def __start_stream_writer(ds_writer: DataStreamWriter, job: SparkStreamJob, streaming_mode: str, batch_interval: str):
    checkpoint_location = get_checkpoint_location(job.output)
    ds_writer = ds_writer.option("checkpointLocation", checkpoint_location)
    streaming_modes = ["once", "microbatch"]
    if streaming_mode == "microbatch":
        if batch_interval is None:
            raise ValueError("batch_interval is required for the microbatch mode")
        else:
            ds_writer = ds_writer.trigger(processingTime=batch_interval)
    elif streaming_mode == "once":
        ds_writer = ds_writer.trigger(once=True)
    else:
        raise ValueError(
            f"value streaming_mode: {streaming_mode} is not one of the supported streaming modes : {streaming_modes}"
        )
    ds_writer.start().awaitTermination()



import os
import uuid
from typing import Dict, Optional

from pyspark.sql import DataFrame

from karadoc.common import conf
from karadoc.common.utils.assert_utils import assert_true


def stream_to_batch(stream_df: DataFrame) -> DataFrame:
    """Transform a streaming dataframe to a batch dataframe.

    :param stream_df: A Spark streaming dataframe
    :return: batch dataframe
    """
    assert_true(stream_df.isStreaming, "dataframe parameter must be a streaming dataframe")
    random_name = "df_" + str(uuid.uuid4()).replace("-", "_")
    stream_df.writeStream.format("memory").queryName(f"df_{random_name}").trigger(once=True).start().awaitTermination()
    df = stream_df.sql_ctx.sparkSession.sql(f"""SELECT * FROM df_{random_name}""")  # nosec B608
    return df


def batch_to_stream(
    df: DataFrame,
    mode: str = "OVERWRITE",
    relative_tmp_path: Optional[str] = None,
    options: Optional[Dict[str, str]] = None,
) -> DataFrame:
    """Transform a batch dataframe to a streaming dataframe.

    :param df: A Spark dataframe
    :param mode: Specifies the behavior when data or table already exists.
        * `append`: Append stream_df dataframe to existing data in the relative_tmp_path.
        * `overwrite`: Overwrite existing data.
        * `error` or `errorifexists`: Throw an exception if data already exists.
        * `ignore`: Silently ignore this operation if data already exists.

    :param relative_tmp_path: A relative tmp path will be used to store temporarily the spark df
    :param options: Specifies the behavior when data or table already exists.
    :return: streaming dataframe
    """
    if options is None:
        options = {}
    assert_true(not df.isStreaming, "dataframe parameter must be a batch dataframe")
    if relative_tmp_path is None:
        relative_tmp_path = str(uuid.uuid4()).replace("-", "_")
    relative_tmp_path = os.path.join(conf.get_spark_stream_tmp_dir(), relative_tmp_path)
    df.write.mode(mode).parquet(relative_tmp_path)
    ds = df.sql_ctx.sparkSession.readStream.options(**options).schema(df.schema).parquet(relative_tmp_path)
    return ds



import adal
from pyspark.sql import DataFrame, SparkSession

from karadoc.common.conf import ConfBox
from karadoc.common.connector import Connector

JDBC_SPARK = "com.microsoft.sqlserver.jdbc.spark"


class MssqlConnector(Connector):
    def __init__(self, spark: SparkSession, connection_conf: ConfBox):
        super().__init__(spark, connection_conf)

    def _get_url(self):
        return "jdbc:sqlserver://%s;databaseName=%s;" % (self.conf.get("host"), self.conf.get("database"))

    def _get_access_token(self):
        client_id = self.conf.get("client_id")
        client_secret = self.conf.get("client_secret")
        tenant_id = self.conf.get("tenant_id")
        authority = "https://login.microsoftonline.com/%s" % tenant_id
        context = adal.AuthenticationContext(authority)
        token = context.acquire_token_with_client_credentials("https://database.windows.net", client_id, client_secret)
        return token["accessToken"]

    def _get_read_options(self, source: dict):
        url = self._get_url()
        access_token = self._get_access_token()
        query = source.get("query")
        table_name = source.get("table")

        read_options = {
            "url": url,
            "accessToken": access_token,
            "encrypt": "true",
            "hostNameInCertificate": "*.database.windows.net",
            "isolationLevel": "READ_UNCOMMITTED",
        }

        if query is not None:
            read_options["query"] = query
        else:
            read_options["dbtable"] = table_name

        read_options.update(source.get("options", {}))
        return read_options

    def _get_write_options(self, dest: dict):
        url = self._get_url()
        access_token = self._get_access_token()
        table_name = dest.get("table")

        write_options = {
            "url": url,
            "dbtable": table_name,
            "accessToken": access_token,
            "encrypt": "true",
            "hostNameInCertificate": "*.database.windows.net",
        }
        write_options.update(dest.get("options", {}))
        return write_options

    def read(self, source: dict):
        """Read data from an external input via a spark connector for SQL Server.

        If the input table has columns that contains dots ('.'), these are
        automatically replaced by underscores to avoid bugs downstream.

        The source must have the following attributes:

        - "connection": name of the jdbc connection
        - "table": name of the table
        - "query": an optional custom query that can be used instead of "table".
        - "options": (Optional[Dict[str, str]]) extra options to pass to the spark connector read method
        # https://github.com/microsoft/sql-spark-connector#supported-options

        The corresponding mssql connection must have the following attributes:

        - type = 'mssql'
        - host = Server name
        - database = Database name
        - client_id = 'client-id'
        - client_secret = 'client-secret'
        - tenant_id = 'tenant-id'

        :param source: a dictionary describing the source
        :return: a DataFrame
        """
        options = self._get_read_options(source)
        df = self.spark.read.format(JDBC_SPARK).options(**options).load()
        # Replace dots with underscore in column names
        df = df.toDF(*[c.replace(".", "_") for c in df.columns])
        return df

    def write(self, df: DataFrame, dest) -> None:
        """Write dataframe to an external output via a jdbc connection.

        The dest must have the following attributes:
        - "connection": name of the jdbc connection
        - "table": name of the table
        - "options": (optional) extra options to pass to the spark connector write method
        # https://github.com/microsoft/sql-spark-connector#supported-options

        The corresponding jdbc connection must have the following attributes:

        - type = 'mssql'
        - host = Server name
        - database = Database name
        - client_id = 'client-id'
        - client_secret = 'client-secret'
        - tenant_id = 'tenant-id'

        :param df: the DataFrame to export
        :param dest: A dictionary describing the destination
        :return:
        """
        mode = dest["mode"]
        options = self._get_write_options(dest)
        df.write.format(JDBC_SPARK).mode(mode).options(**options).save()

    def test_connection(self):
        options = self._get_read_options({"query": "SELECT 1 as ok"})
        df = self.spark.read.format(JDBC_SPARK).options(**options).load()
        if df.collect()[0].ok == 1:
            return True
        else:
            df.show()
            return False



from io import StringIO
from typing import TYPE_CHECKING, Dict, Optional, Union

from pyspark.sql import DataFrame

if TYPE_CHECKING:
    import pandas as pd

default_encoding = "UTF-8"


def write_pandas_dataframes_to_xlsx(
    dfs: Union["pd.DataFrame", Dict[str, "pd.DataFrame"]], output: str, **options
) -> None:
    """Writes one or multiple DataFrames (as dicts) in an Excel file (with one sheet per DataFrame if multiple
    DataFrames were given).

    Extra options will be passed through to the
    `pandas.DataFrame.to_excel()
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_excel.html>`_ method

    Code has been shamelessly pumped from `StackOverflow
    <https://stackoverflow.com/questions/17326973/is-there-a-way-to-auto-adjust-excel-column-widths-with-pandas-excelwriter>`_

    Known Limitations
    -----------------
    When exporting multiple DataFrames, the sheet name length must not exceed 31 characters.

    :param dfs: A Pandas DataFrame or a Dict of multiple (sheet_name, DataFrame)
    :param output: the path of the output excel file
    :param options: extra options to pass to `pandas.DataFrame.to_excel()`
    :return:
    """
    import pandas as pd

    if isinstance(dfs, pd.DataFrame):
        dfs = {"Sheet1": dfs}
    writer = pd.ExcelWriter(output, engine="xlsxwriter")
    for sheetname, df in dfs.items():  # loop through `dict` of dataframes
        df.to_excel(writer, sheet_name=sheetname, **options)  # send df to writer
        worksheet = writer.sheets[sheetname]  # pull worksheet object
        for idx, col in enumerate(df):  # loop through all columns
            series = df[col]
            max_len = (
                max(
                    (
                        series.astype(str).map(len).max(),  # len of largest item
                        len(str(series.name)),  # len of column name/header
                    )
                )
                + 1
            )  # adding a little extra space
            worksheet.set_column(idx, idx, max_len)  # set column width
    writer.save()


def __export_string(string: str, output: Optional[str]) -> None:
    if output is None:
        print(string)
    else:
        with open(output, "w") as f:
            f.write(string)


def local_export_dataframe(df: DataFrame, output: Optional[str], format: str) -> None:
    """Export a Spark DataFrame on the local filesystem with the given output and format.

    Supported formats are:
    - markdown
    - csv
    - xlsx

    If no output is specified, the result will be directly printed on stdout, except for binary formats (e.g. xlsx)
    which requires an output path.

    :param df: a Spark DataFrame
    :param output: (Optional) the output file path
    :param format: Either one of ["markdown", "csv", "xlsx"]
    :return:
    """
    pdf = df.toPandas()
    if format == "xlsx":
        if output is None:
            raise ValueError(f"The output file path is required when exporting to binary format {format}")
        write_pandas_dataframes_to_xlsx(pdf, output, index=False, freeze_panes=(1, 0))
        print(f"Output written to file {output}")
    elif format == "csv":
        if output is None:
            buffer = StringIO()
            pdf.to_csv(buffer, index=False)
            print(buffer.getvalue().encode(default_encoding))
        else:
            pdf.to_csv(output, index=False, encoding=default_encoding)
            print(f"Output written to file {output}")
    elif format == "markdown":
        string = pdf.to_markdown(index=False)
        __export_string(string, output)
    elif format == "txt":
        string = "\n".join([row[0] for row in df.collect()])
        __export_string(string, output)
    else:
        raise ValueError(f"Unknown format: {format}")


def local_export_dataframes_xlsx(dataframes: Dict[str, DataFrame], output: str) -> None:
    """Export multiple Spark DataFrames on the local filesystem as an Excel file, with one tab per DataFrame.

    Supported formats are:
    - markdown
    - csv
    - xlsx

    If no output is specified, the result will be directly printed on stdout, except for binary formats (e.g. xlsx)
    which requires an output path.

    :param df: a Spark DataFrame
    :param output: (Optional) the output file path
    :param format: Either one of ["markdown", "csv", "xlsx"]
    :return:
    """
    p_dataframes = {k: v.toPandas() for k, v, in dataframes.items()}
    if output is None:
        raise ValueError("The output file path is required when exporting to binary format xlsx")
    write_pandas_dataframes_to_xlsx(p_dataframes, output, index=False, freeze_panes=(1, 0))
    print(f"Output written to file {output}")



from argparse import ArgumentParser, Namespace
from typing import Optional

from pyspark.sql import DataFrame

from karadoc.common import Job, conf, run
from karadoc.common.commands.command import Command
from karadoc.common.commands.help import LineBreakHelpFormatter
from karadoc.common.commands.options.dry_option import DryOption
from karadoc.common.commands.options.read_from_option import ReadFromOption
from karadoc.common.commands.options.spark_conf_option import SparkConfOption
from karadoc.common.commands.options.tables_option import TablesOption
from karadoc.common.commands.options.vars_option import VarsOption
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.commands.spark import init_job
from karadoc.common.exceptions import JobDisabledException
from karadoc.common.job_core.has_spark import _partition_to_path
from karadoc.common.model import variables
from karadoc.common.nonreg import DataframeComparator, NonregResult
from karadoc.common.run.spark_batch_job import SparkBatchJob
from karadoc.common.validations import fail_if_results
from karadoc.common.validations.connection_validations import validate_connections
from karadoc.common.validations.job_validations import (
    validate_inputs,
    validate_output_partition,
)


def _get_join_cols(args, job):
    if args.nonreg_join_cols:
        join_cols = list(args.nonreg_join_cols.split(","))
    elif job.standardized_primary_key is not None:
        join_cols = list(job.standardized_primary_key)
    else:
        join_cols = None
    return join_cols


def _check_nonreg(job: Job, df: DataFrame, args):
    """Compare the output of a Populate with the currently existing one."""
    if conf.is_dev_env():
        job.spark.conf.set("spark.sql.shuffle.partitions", "200")

    previous_df = get_previous_df(job=job, compare_env=args.compare_env)

    if df is None or not isinstance(df, DataFrame):
        print("ERROR: The job did not return any DataFrame.")
        return ReturnCode.Error

    if len(df.columns) == 0:
        print("ERROR: The job returned a DataFrame with zero columns.")
        return ReturnCode.Error

    join_cols = _get_join_cols(args, job)

    if args.nonreg_filter:
        df = df.filter(args.nonreg_filter)
        previous_df = previous_df.filter(args.nonreg_filter)

    ignore_cols = args.ignore_cols
    if args.ignore_cols is not None:
        previous_df = previous_df.drop(*ignore_cols)
        df = df.drop(*ignore_cols)

    from pyspark.storagelevel import StorageLevel

    df.persist(StorageLevel.DISK_ONLY)
    nonreg = DataframeComparator(nb_diffed_rows=args.nonreg_nb_rows, left_df_alias="before", right_df_alias="after")
    nonreg_result = nonreg.compare_df(previous_df, df, join_cols=join_cols, show_examples=args.show_examples)
    inspect_nonreg_results(nonreg_result)
    if nonreg_result.is_ok:
        return ReturnCode.Success
    else:
        return ReturnCode.Error


def get_previous_df(job, compare_env):
    if compare_env is not None:
        partition_path = _partition_to_path(job.output_partition)
        output = f"{job.output}{partition_path}"
        previous_df_hdfs_location = job.hdfs_input(output, remote_env_name=compare_env)
    else:
        previous_df_hdfs_location = job.hdfs_output()
    previous_df = job._read_path(previous_df_hdfs_location, job.output_format)
    return previous_df


def inspect_nonreg_results(df: NonregResult):
    """Function used for unit tests. Mock this function to be able to inspect the produced DataFrame
    during the execution of the command."""
    pass


def _run_table(job: SparkBatchJob, args) -> Optional[DataFrame]:
    """Computes the Spark DataFrame build by a SparkBatchJob with the specified args.

    :param job:
    :param args:
    :return:
    """
    df: Optional[DataFrame] = None
    if args.remote_env is not None:
        job._configure_remote_input(args.remote_env)
    run.print_job_properties(job)
    fail_if_results(validate_inputs(job))
    fail_if_results(validate_output_partition(job))
    fail_if_results(validate_connections(job))
    if job.disable:
        if not args.dry:
            raise JobDisabledException()
        else:
            print("WARN: The job has been disabled, it should not be launched")
    if not args.dry:
        init_job(job=job, raw_args=args.raw_args, spark_conf=args.spark_conf)
        df = job.run()
    elif not conf.is_dev_env():
        init_job(job=job, raw_args=args.raw_args, spark_conf=args.spark_conf)
    return df


class NonregCommand(Command):
    description = "spot differences for the specified tables"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.formatter_class = LineBreakHelpFormatter

        TablesOption.add_arguments(parser)
        DryOption.add_arguments(parser)
        ReadFromOption.add_arguments(parser)
        VarsOption.add_arguments(parser)
        SparkConfOption.add_arguments(parser)
        parser.add_argument(
            "--filter",
            dest="nonreg_filter",
            type=str,
            help="Filter to apply to each DataFrame before comparing them."
            " (works only with the --nonreg option)"
            """\nExample: --nonreg-filter "day > '2020-07-25'" """,
        )
        parser.add_argument(
            "--nb-rows",
            dest="nonreg_nb_rows",
            type=int,
            default=10,
            help="Number of differing rows to show when performing a non-reg (works only with the --nonreg option)",
        )
        parser.add_argument(
            "--join-cols",
            dest="nonreg_join_cols",
            type=str,
            help="Comma-separated list of columns names to use to perform the nonreg diff"
            " (works only with the --nonreg option)",
        )
        parser.add_argument(
            "--compare-with",
            dest="compare_env",
            metavar="ENV",
            help="Compare with data from a remote warehouse in env ENV instead of the local warehouse",
        )
        parser.add_argument(
            "--ignore-cols", dest="ignore_cols", type=str, nargs="*", help="Columns specified will be ignored"
        )
        parser.add_argument(
            "--show-examples",
            dest="show_examples",
            default=False,
            action="store_true",
            help="If set, print for each column examples of full rows where this column changes",
        )

    @staticmethod
    def do_command(args: Namespace) -> ReturnCode:
        return_code = ReturnCode.Success
        vars_list = variables.expand_vars(args.vars)
        for table in args.tables:
            for vars in vars_list:
                print("\nStarting table %s" % table)
                job: SparkBatchJob = run.load_runnable_populate(table, vars)
                df = _run_table(job, args)
                if df is not None:
                    return_code = _check_nonreg(job, df, args)
        return return_code



from typing import Dict, Iterator, Optional

from pyspark.sql import DataFrame

from karadoc.common import conf
from karadoc.common.conf import CONNECTION_GROUP
from karadoc.common.model import file_index
from karadoc.common.run import load_populate
from karadoc.common.stream import load_stream_file
from karadoc.common.table_utils import populate_exists, stream_file_exists


def describe_connection(job, connection_name, source_details, full_table_name, direction, env: Optional[str]):
    try:
        conn = conf.get_connection_conf(connection_name, env)
        common_args = {
            "job_type": type(job).__name__,
            "full_table_name": full_table_name,
            "conn_name": connection_name,
            "conn_type": conn["type"],
            "direction": direction,
            "job_disabled": str(job.disable),
            "conn_disabled": str(conn.get("disable", False)),
        }
        if conn["type"].endswith(".jdbc"):
            return {
                **common_args,
                "host": conn["host"],
                "username": conn["user"],
                "database": conn["database"],
                "table": source_details.get("table"),
            }
        elif conn["type"].endswith(".mssql"):
            return {
                **common_args,
                "host": conn["host"],
                "database": conn["database"],
                "table": source_details.get("table"),
            }
        elif conn["type"].endswith(".cosmosdb"):
            return {
                **common_args,
                "host": conn["host"],
                "database": source_details.get("database"),
                "table": source_details.get("collection"),
            }
        elif conn["type"].endswith(".wasbs"):
            return {
                **common_args,
                "host": conn["storage_account"],
                "container": conn["container_name"],
                "table": source_details.get("path"),
            }
        elif conn["type"].endswith(".azure_table"):
            return {
                **common_args,
                "host": conn["storage_account"],
                "table": source_details.get("table"),
            }
        elif conn["type"].endswith(".event_hub"):
            return {
                **common_args,
                "host": conn["HubName"],
                "username": source_details.get("consumer_group"),
                "table": source_details.get("entity_path"),
            }
        elif conn["type"].endswith(".sftp"):
            return {
                **common_args,
                "host": conn["host"],
                "username": conn["username"],
            }
        elif conn["type"].endswith(".bigquery") or conn["type"].endswith(".bigquery_2"):
            return {
                **common_args,
                "container": conn["bucket_name"],
                "database": conn["project_name"],
                "table": source_details.get("table"),
            }
        else:
            return common_args
    except Exception as e:
        return {
            "full_table_name": full_table_name,
            "conn_name": connection_name,
            "conn_type": "error",
            "direction": direction,
            "host": str(e),
            "disabled": str(job.disable),
        }


def __build_connections_generator(env: Optional[str]) -> Iterator[Dict[str, str]]:  # NOSONAR
    # TODO: refactor this method by making a generic load_action_files method
    for (schema, table, path) in list(file_index.list_schema_table_folders()):
        full_table_name = "%s.%s" % (schema, table)
        if populate_exists(full_table_name):
            job = load_populate(full_table_name)
            for source_name, source_details in job.external_inputs.items():
                connection_name = source_details[CONNECTION_GROUP]
                yield describe_connection(
                    job, connection_name, source_details, full_table_name, direction="input", env=env
                )
            for dest_name, dest_details in job.external_outputs.items():
                connection_name = dest_details[CONNECTION_GROUP]
                yield describe_connection(
                    job, connection_name, dest_details, full_table_name, direction="output", env=env
                )
        if stream_file_exists(full_table_name):
            job = load_stream_file(full_table_name)
            for source_name, source_details in job.external_inputs.items():
                connection_name = source_details[CONNECTION_GROUP]
                yield describe_connection(
                    job, connection_name, source_details, full_table_name, direction="input", env=env
                )
            if job.external_output is not None:
                yield describe_connection(
                    job,
                    job.external_output[CONNECTION_GROUP],
                    job.external_output,
                    full_table_name,
                    direction="output",
                    env=env,
                )


def list_connections_df(env: Optional[str]) -> DataFrame:
    """Builds a Spark DataFrame listing all external inputs/outputs used by POPULATE.py files, with their respective
    source/dest and connection properties.

    :param env: Retrieves the connections properties for the specified environment
    :return: a Spark DataFrame
    """
    from karadoc.common import Job

    job = Job()
    job.init()
    data = __build_connections_generator(env)
    schema = """
        full_table_name: STRING,
        job_type: STRING,
        conn_name: STRING,
        conn_type: STRING,
        direction: STRING,
        host: STRING,
        username: STRING,
        container: STRING,
        database: STRING,
        table: STRING,
        job_disabled: STRING,
        conn_disabled: STRING
    """
    df = job.spark.createDataFrame(data, schema=schema)
    df = df.orderBy("full_table_name")
    return df



from typing import List


class MissingConfigurationException(Exception):
    """Exception raised when we fail to load a configuration that has no default value"""

    def __init__(self, conf_id: str) -> None:
        msg = f"Could not find configuration {conf_id}\n"
        "Please make sure your environment variable KARADOC_ENV is correctly set"
        Exception.__init__(self, msg)


class MissingConfigurationParameterException(Exception):
    """Exception raised when a required parameter is missing from a ConfigurableClass"""

    def __init__(self, param_name: str) -> None:
        msg = f"Missing required parameter {param_name}"
        Exception.__init__(self, msg)


class MissingVariablesException(Exception):
    """Exception raised when a job requiring variables is run without specifying variable values"""

    def __init__(self, missing_keys: List[str]) -> None:
        missing_keys_str = ", ".join(missing_keys)
        msg = f"The following variables are required and missing: [{missing_keys_str}]"
        Exception.__init__(self, msg)


class JobDisabledException(Exception):
    def __init__(self) -> None:
        msg = "The job has been disabled, it should not be launched"
        Exception.__init__(self, msg)


class ActionFileLoadingError(Exception):
    """Error raised when failing to laod an Action File"""

    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)


class CommandValidationError(Exception):
    """Error raised when a command fails validation"""

    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)


class ConnectorLoadError(Exception):
    """Error raised when a failing to load a connector"""

    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)


class LoggedException(Exception):
    """Exception raised when an exception has already been caught and logged"""

    def __init__(self) -> None:
        Exception.__init__(self, "See parent exception")


class IllegalJobInitError(Exception):
    """Error raised when job.init() is called at an inappropriate time"""


class ForbiddenActionError(Exception):
    """Error raised when something that should not happen does happen"""
